"""
    pygments.lexers.webassembly
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Lexers for the WebAssembly text format.

    The grammar can be found at https://github.com/WebAssembly/spec/blob/master/interpreter/README.md
    and https://webassembly.github.io/spec/core/text/.


    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, words, bygroups, default
from pygments.token import Text, Comment, Operator, Keyword, String, Number, Punctuation, Name

__all__ = ['WatLexer']

keywords = (
    'module', 'import', 'func', 'funcref', 'start', 'param', 'local', 'type',
    'result', 'export', 'memory', 'global', 'mut', 'data', 'table', 'elem',
    'if', 'then', 'else', 'end', 'block', 'loop'
)

builtins = (
    'unreachable', 'nop', 'block', 'loop', 'if', 'else', 'end', 'br', 'br_if',
    'br_table', 'return', 'call', 'call_indirect', 'drop', 'select',
    'local.get', 'local.set', 'local.tee', 'global.get', 'global.set',
    'i32.load', 'i64.load', 'f32.load', 'f64.load', 'i32.load8_s',
    'i32.load8_u', 'i32.load16_s', 'i32.load16_u', 'i64.load8_s',
    'i64.load8_u', 'i64.load16_s', 'i64.load16_u', 'i64.load32_s',
    'i64.load32_u', 'i32.store', 'i64.store', 'f32.store', 'f64.store',
    'i32.store8', 'i32.store16', 'i64.store8', 'i64.store16', 'i64.store32',
    'memory.size', 'memory.grow', 'i32.const', 'i64.const', 'f32.const',
    'f64.const', 'i32.eqz', 'i32.eq', 'i32.ne', 'i32.lt_s', 'i32.lt_u',
    'i32.gt_s', 'i32.gt_u', 'i32.le_s', 'i32.le_u', 'i32.ge_s', 'i32.ge_u',
    'i64.eqz', 'i64.eq', 'i64.ne', 'i64.lt_s', 'i64.lt_u', 'i64.gt_s',
    'i64.gt_u', 'i64.le_s', 'i64.le_u', 'i64.ge_s', 'i64.ge_u', 'f32.eq',
    'f32.ne', 'f32.lt', 'f32.gt', 'f32.le', 'f32.ge', 'f64.eq', 'f64.ne',
    'f64.lt', 'f64.gt', 'f64.le', 'f64.ge', 'i32.clz', 'i32.ctz', 'i32.popcnt',
    'i32.add', 'i32.sub', 'i32.mul', 'i32.div_s', 'i32.div_u', 'i32.rem_s',
    'i32.rem_u', 'i32.and', 'i32.or', 'i32.xor', 'i32.shl', 'i32.shr_s',
    'i32.shr_u', 'i32.rotl', 'i32.rotr', 'i64.clz', 'i64.ctz', 'i64.popcnt',
    'i64.add', 'i64.sub', 'i64.mul', 'i64.div_s', 'i64.div_u', 'i64.rem_s',
    'i64.rem_u', 'i64.and', 'i64.or', 'i64.xor', 'i64.shl', 'i64.shr_s',
    'i64.shr_u', 'i64.rotl', 'i64.rotr', 'f32.abs', 'f32.neg', 'f32.ceil',
    'f32.floor', 'f32.trunc', 'f32.nearest', 'f32.sqrt', 'f32.add', 'f32.sub',
    'f32.mul', 'f32.div', 'f32.min', 'f32.max', 'f32.copysign', 'f64.abs',
    'f64.neg', 'f64.ceil', 'f64.floor', 'f64.trunc', 'f64.nearest', 'f64.sqrt',
    'f64.add', 'f64.sub', 'f64.mul', 'f64.div', 'f64.min', 'f64.max',
    'f64.copysign', 'i32.wrap_i64', 'i32.trunc_f32_s', 'i32.trunc_f32_u',
    'i32.trunc_f64_s', 'i32.trunc_f64_u', 'i64.extend_i32_s',
    'i64.extend_i32_u', 'i64.trunc_f32_s', 'i64.trunc_f32_u',
    'i64.trunc_f64_s', 'i64.trunc_f64_u', 'f32.convert_i32_s',
    'f32.convert_i32_u', 'f32.convert_i64_s', 'f32.convert_i64_u',
    'f32.demote_f64', 'f64.convert_i32_s', 'f64.convert_i32_u',
    'f64.convert_i64_s', 'f64.convert_i64_u', 'f64.promote_f32',
    'i32.reinterpret_f32', 'i64.reinterpret_f64', 'f32.reinterpret_i32',
    'f64.reinterpret_i64',
)


class WatLexer(RegexLexer):
    """Lexer for the WebAssembly text format.

    .. versionadded:: 2.9
    """

    name = 'WebAssembly'
    url = 'https://webassembly.org/'
    aliases = ['wast', 'wat']
    filenames = ['*.wat', '*.wast']

    tokens = {
        'root': [
            (words(keywords, suffix=r'(?=[^a-z_\.])'), Keyword),
            (words(builtins), Name.Builtin, 'arguments'),
            (words(['i32', 'i64', 'f32', 'f64']), Keyword.Type),
            (r'\$[A-Za-z0-9!#$%&\'*+./:<=>?@\\^_`|~-]+', Name.Variable), # yes, all of the are valid in identifiers
            (r';;.*?$', Comment.Single),
            (r'\(;', Comment.Multiline, 'nesting_comment'),
            (r'[+-]?0x[\dA-Fa-f](_?[\dA-Fa-f])*(.([\dA-Fa-f](_?[\dA-Fa-f])*)?)?([pP][+-]?[\dA-Fa-f](_?[\dA-Fa-f])*)?', Number.Float),
            (r'[+-]?\d.\d(_?\d)*[eE][+-]?\d(_?\d)*', Number.Float),
            (r'[+-]?\d.\d(_?\d)*', Number.Float),
            (r'[+-]?\d.[eE][+-]?\d(_?\d)*', Number.Float),
            (r'[+-]?(inf|nan:0x[\dA-Fa-f](_?[\dA-Fa-f])*|nan)', Number.Float),
            (r'[+-]?0x[\dA-Fa-f](_?[\dA-Fa-f])*', Number.Hex),
            (r'[+-]?\d(_?\d)*', Number.Integer),
            (r'[\(\)]', Punctuation),
            (r'"', String.Double, 'string'),
            (r'\s+', Text),
        ],
        'nesting_comment': [
            (r'\(;', Comment.Multiline, '#push'),
            (r';\)', Comment.Multiline, '#pop'),
            (r'[^;(]+', Comment.Multiline),
            (r'[;(]', Comment.Multiline),
        ],
        'string': [
            (r'\\[\dA-Fa-f][\dA-Fa-f]', String.Escape), # must have exactly two hex digits
            (r'\\t', String.Escape),
            (r'\\n', String.Escape),
            (r'\\r', String.Escape),
            (r'\\"', String.Escape),
            (r"\\'", String.Escape),
            (r'\\u\{[\dA-Fa-f](_?[\dA-Fa-f])*\}', String.Escape),
            (r'\\\\', String.Escape),
            (r'"', String.Double, '#pop'),
            (r'[^"\\]+', String.Double),
        ],
        'arguments': [
            (r'\s+', Text),
            (r'(offset)(=)(0x[\dA-Fa-f](_?[\dA-Fa-f])*)', bygroups(Keyword, Operator, Number.Hex)),
            (r'(offset)(=)(\d(_?\d)*)', bygroups(Keyword, Operator, Number.Integer)),
            (r'(align)(=)(0x[\dA-Fa-f](_?[\dA-Fa-f])*)', bygroups(Keyword, Operator, Number.Hex)),
            (r'(align)(=)(\d(_?\d)*)', bygroups(Keyword, Operator, Number.Integer)),
            default('#pop'),
        ]
    }



# Engines
from testrail_api_reporter.engines.at_coverage_reporter import ATCoverageReporter
from testrail_api_reporter.engines.results_reporter import TestRailResultsReporter
from testrail_api_reporter.engines.plotly_reporter import PlotlyReporter
from testrail_api_reporter.engines.case_backup import TCBackup
# Publishers
from testrail_api_reporter.publishers.confluence_sender import ConfluenceSender
from testrail_api_reporter.publishers.email_sender import EmailSender
from testrail_api_reporter.publishers.slack_sender import SlackSender
from testrail_api_reporter.publishers.gdrive_uploader import GoogleDriveUploader
# Utils
from testrail_api_reporter.utils.reporter_utils import upload_image, zip_file, delete_file



"""
    pygments.lexers.macaulay2
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Lexer for Macaulay2.

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, words
from pygments.token import Comment, Keyword, Name, String, Text

__all__ = ['Macaulay2Lexer']

# Auto-generated for Macaulay2-1.22. Do not modify this file manually.

M2KEYWORDS = (
    "and",
    "break",
    "catch",
    "continue",
    "do",
    "elapsedTime",
    "elapsedTiming",
    "else",
    "for",
    "from",
    "global",
    "if",
    "in",
    "list",
    "local",
    "new",
    "not",
    "of",
    "or",
    "return",
    "shield",
    "SPACE",
    "step",
    "symbol",
    "then",
    "threadVariable",
    "throw",
    "time",
    "timing",
    "to",
    "try",
    "when",
    "while",
    "xor"
    )

M2DATATYPES = (
    "Adjacent",
    "AffineVariety",
    "Analyzer",
    "ANCHOR",
    "AngleBarList",
    "Array",
    "AssociativeExpression",
    "Bag",
    "BasicList",
    "BettiTally",
    "BinaryOperation",
    "BLOCKQUOTE",
    "BODY",
    "BOLD",
    "Boolean",
    "BR",
    "BUTTON",
    "CacheFunction",
    "CacheTable",
    "CC",
    "CDATA",
    "ChainComplex",
    "ChainComplexMap",
    "CODE",
    "CoherentSheaf",
    "Command",
    "COMMENT",
    "CompiledFunction",
    "CompiledFunctionBody",
    "CompiledFunctionClosure",
    "ComplexField",
    "Constant",
    "Database",
    "DD",
    "Descent",
    "Describe",
    "Dictionary",
    "DirectSum",
    "DIV",
    "Divide",
    "DL",
    "DocumentTag",
    "DT",
    "Eliminate",
    "EM",
    "EngineRing",
    "Equation",
    "ExampleItem",
    "Expression",
    "File",
    "FilePosition",
    "FractionField",
    "Function",
    "FunctionApplication",
    "FunctionBody",
    "FunctionClosure",
    "GaloisField",
    "GeneralOrderedMonoid",
    "GlobalDictionary",
    "GradedModule",
    "GradedModuleMap",
    "GroebnerBasis",
    "GroebnerBasisOptions",
    "HashTable",
    "HEAD",
    "HEADER1",
    "HEADER2",
    "HEADER3",
    "HEADER4",
    "HEADER5",
    "HEADER6",
    "HeaderType",
    "Holder",
    "HR",
    "HREF",
    "HTML",
    "Hybrid",
    "Hypertext",
    "HypertextContainer",
    "HypertextParagraph",
    "HypertextVoid",
    "Ideal",
    "IMG",
    "ImmutableType",
    "INDENT",
    "IndeterminateNumber",
    "IndexedVariable",
    "IndexedVariableTable",
    "InexactField",
    "InexactFieldFamily",
    "InexactNumber",
    "InfiniteNumber",
    "INPUT",
    "IntermediateMarkUpType",
    "ITALIC",
    "Iterator",
    "KBD",
    "Keyword",
    "LABEL",
    "LATER",
    "LI",
    "LINK",
    "List",
    "LITERAL",
    "LocalDictionary",
    "LowerBound",
    "Manipulator",
    "MapExpression",
    "MarkUpType",
    "Matrix",
    "MatrixExpression",
    "MENU",
    "META",
    "MethodFunction",
    "MethodFunctionBinary",
    "MethodFunctionSingle",
    "MethodFunctionWithOptions",
    "Minus",
    "Module",
    "Monoid",
    "MonoidElement",
    "MonomialIdeal",
    "MultigradedBettiTally",
    "MutableHashTable",
    "MutableList",
    "MutableMatrix",
    "Net",
    "NetFile",
    "Nothing",
    "Number",
    "NumberedVerticalList",
    "OL",
    "OneExpression",
    "Option",
    "OptionTable",
    "OrderedMonoid",
    "Package",
    "PARA",
    "Parenthesize",
    "Parser",
    "Partition",
    "PolynomialRing",
    "Power",
    "PRE",
    "Product",
    "ProductOrder",
    "Program",
    "ProgramRun",
    "ProjectiveHilbertPolynomial",
    "ProjectiveVariety",
    "Pseudocode",
    "QQ",
    "QuotientRing",
    "RealField",
    "Resolution",
    "Ring",
    "RingElement",
    "RingFamily",
    "RingMap",
    "RowExpression",
    "RR",
    "RRi",
    "SAMP",
    "SCRIPT",
    "ScriptedFunctor",
    "SelfInitializingType",
    "Sequence",
    "Set",
    "SheafExpression",
    "SheafOfRings",
    "SMALL",
    "SPAN",
    "SparseMonomialVectorExpression",
    "SparseVectorExpression",
    "String",
    "STRONG",
    "STYLE",
    "SUB",
    "Subscript",
    "SUBSECTION",
    "Sum",
    "SumOfTwists",
    "SUP",
    "Superscript",
    "Symbol",
    "SymbolBody",
    "TABLE",
    "Table",
    "Tally",
    "Task",
    "TD",
    "TensorProduct",
    "TestInput",
    "TEX",
    "TH",
    "Thing",
    "Time",
    "TITLE",
    "TO",
    "TO2",
    "TOH",
    "TR",
    "TT",
    "Type",
    "UL",
    "URL",
    "VAR",
    "Variety",
    "Vector",
    "VectorExpression",
    "VerticalList",
    "VirtualTally",
    "VisibleList",
    "WrapperType",
    "ZeroExpression",
    "ZZ"
    )

M2FUNCTIONS = (
    "about",
    "abs",
    "accumulate",
    "acos",
    "acosh",
    "acot",
    "acoth",
    "addCancelTask",
    "addDependencyTask",
    "addEndFunction",
    "addHook",
    "addStartFunction",
    "addStartTask",
    "adjoint",
    "agm",
    "alarm",
    "all",
    "ambient",
    "analyticSpread",
    "ancestor",
    "ancestors",
    "andP",
    "ann",
    "annihilator",
    "antipode",
    "any",
    "append",
    "applicationDirectory",
    "apply",
    "applyKeys",
    "applyPairs",
    "applyTable",
    "applyValues",
    "apropos",
    "arXiv",
    "ascii",
    "asin",
    "asinh",
    "ass",
    "assert",
    "associatedGradedRing",
    "associatedPrimes",
    "atan",
    "atan2",
    "atanh",
    "atEndOfFile",
    "autoload",
    "baseFilename",
    "baseName",
    "baseRing",
    "basis",
    "beginDocumentation",
    "benchmark",
    "BesselJ",
    "BesselY",
    "Beta",
    "betti",
    "between",
    "binomial",
    "borel",
    "cacheValue",
    "cancelTask",
    "capture",
    "ceiling",
    "centerString",
    "chainComplex",
    "changeBase",
    "char",
    "characters",
    "charAnalyzer",
    "check",
    "checkDegrees",
    "chi",
    "class",
    "clean",
    "clearEcho",
    "code",
    "codim",
    "coefficient",
    "coefficientRing",
    "coefficients",
    "cohomology",
    "coimage",
    "coker",
    "cokernel",
    "collectGarbage",
    "columnAdd",
    "columnate",
    "columnMult",
    "columnPermute",
    "columnRankProfile",
    "columnSwap",
    "combine",
    "commandInterpreter",
    "commonest",
    "commonRing",
    "comodule",
    "complement",
    "complete",
    "components",
    "compose",
    "compositions",
    "compress",
    "concatenate",
    "conductor",
    "cone",
    "conjugate",
    "connectionCount",
    "constParser",
    "content",
    "contract",
    "conwayPolynomial",
    "copy",
    "copyDirectory",
    "copyFile",
    "cos",
    "cosh",
    "cot",
    "cotangentSheaf",
    "coth",
    "cover",
    "coverMap",
    "cpuTime",
    "createTask",
    "csc",
    "csch",
    "currentColumnNumber",
    "currentDirectory",
    "currentPosition",
    "currentRowNumber",
    "currentTime",
    "deadParser",
    "debug",
    "debugError",
    "decompose",
    "deepSplice",
    "default",
    "degree",
    "degreeGroup",
    "degreeLength",
    "degrees",
    "degreesMonoid",
    "degreesRing",
    "delete",
    "demark",
    "denominator",
    "depth",
    "describe",
    "det",
    "determinant",
    "diagonalMatrix",
    "diameter",
    "dictionary",
    "diff",
    "difference",
    "Digamma",
    "dim",
    "directSum",
    "disassemble",
    "discriminant",
    "dismiss",
    "distinguished",
    "divideByVariable",
    "doc",
    "document",
    "drop",
    "dual",
    "eagonNorthcott",
    "echoOff",
    "echoOn",
    "eigenvalues",
    "eigenvectors",
    "eint",
    "elements",
    "eliminate",
    "End",
    "endPackage",
    "entries",
    "erase",
    "erf",
    "erfc",
    "error",
    "euler",
    "eulers",
    "even",
    "EXAMPLE",
    "examples",
    "exec",
    "exp",
    "expectedReesIdeal",
    "expm1",
    "exponents",
    "export",
    "exportFrom",
    "exportMutable",
    "expression",
    "extend",
    "exteriorPower",
    "factor",
    "Fano",
    "fileExecutable",
    "fileExists",
    "fileLength",
    "fileMode",
    "fileReadable",
    "fileTime",
    "fileWritable",
    "fillMatrix",
    "findFiles",
    "findHeft",
    "findProgram",
    "findSynonyms",
    "first",
    "firstkey",
    "fittingIdeal",
    "flagLookup",
    "flatten",
    "flattenRing",
    "flip",
    "floor",
    "fold",
    "forceGB",
    "fork",
    "format",
    "formation",
    "frac",
    "fraction",
    "frames",
    "fromDividedPowers",
    "fromDual",
    "functionBody",
    "futureParser",
    "Gamma",
    "gb",
    "gbRemove",
    "gbSnapshot",
    "gcd",
    "gcdCoefficients",
    "gcdLLL",
    "GCstats",
    "genera",
    "generateAssertions",
    "generator",
    "generators",
    "genericMatrix",
    "genericSkewMatrix",
    "genericSymmetricMatrix",
    "gens",
    "genus",
    "get",
    "getc",
    "getChangeMatrix",
    "getenv",
    "getGlobalSymbol",
    "getNetFile",
    "getNonUnit",
    "getPrimeWithRootOfUnity",
    "getSymbol",
    "getWWW",
    "GF",
    "globalAssign",
    "globalAssignFunction",
    "globalAssignment",
    "globalReleaseFunction",
    "gradedModule",
    "gradedModuleMap",
    "gramm",
    "graphIdeal",
    "graphRing",
    "Grassmannian",
    "groebnerBasis",
    "groupID",
    "hash",
    "hashTable",
    "heft",
    "height",
    "hermite",
    "hilbertFunction",
    "hilbertPolynomial",
    "hilbertSeries",
    "hold",
    "Hom",
    "homogenize",
    "homology",
    "homomorphism",
    "hooks",
    "horizontalJoin",
    "html",
    "httpHeaders",
    "hypertext",
    "icFracP",
    "icFractions",
    "icMap",
    "icPIdeal",
    "ideal",
    "idealizer",
    "identity",
    "image",
    "imaginaryPart",
    "importFrom",
    "independentSets",
    "index",
    "indices",
    "inducedMap",
    "inducesWellDefinedMap",
    "info",
    "input",
    "insert",
    "installAssignmentMethod",
    "installedPackages",
    "installHilbertFunction",
    "installMethod",
    "installMinprimes",
    "installPackage",
    "instance",
    "instances",
    "integralClosure",
    "integrate",
    "intersect",
    "intersectInP",
    "intersection",
    "interval",
    "inverse",
    "inverseErf",
    "inversePermutation",
    "inverseRegularizedBeta",
    "inverseRegularizedGamma",
    "inverseSystem",
    "irreducibleCharacteristicSeries",
    "irreducibleDecomposition",
    "isAffineRing",
    "isANumber",
    "isBorel",
    "isc",
    "isCanceled",
    "isCommutative",
    "isConstant",
    "isDirectory",
    "isDirectSum",
    "isEmpty",
    "isField",
    "isFinite",
    "isFinitePrimeField",
    "isFreeModule",
    "isGlobalSymbol",
    "isHomogeneous",
    "isIdeal",
    "isInfinite",
    "isInjective",
    "isInputFile",
    "isIsomorphic",
    "isIsomorphism",
    "isLinearType",
    "isListener",
    "isLLL",
    "isMember",
    "isModule",
    "isMonomialIdeal",
    "isMutable",
    "isNormal",
    "isOpen",
    "isOutputFile",
    "isPolynomialRing",
    "isPrimary",
    "isPrime",
    "isPrimitive",
    "isPseudoprime",
    "isQuotientModule",
    "isQuotientOf",
    "isQuotientRing",
    "isReady",
    "isReal",
    "isReduction",
    "isRegularFile",
    "isRing",
    "isSkewCommutative",
    "isSorted",
    "isSquareFree",
    "isStandardGradedPolynomialRing",
    "isSubmodule",
    "isSubquotient",
    "isSubset",
    "isSupportedInZeroLocus",
    "isSurjective",
    "isTable",
    "isUnit",
    "isWellDefined",
    "isWeylAlgebra",
    "iterator",
    "jacobian",
    "jacobianDual",
    "join",
    "ker",
    "kernel",
    "kernelLLL",
    "kernelOfLocalization",
    "keys",
    "kill",
    "koszul",
    "last",
    "lcm",
    "leadCoefficient",
    "leadComponent",
    "leadMonomial",
    "leadTerm",
    "left",
    "length",
    "letterParser",
    "lift",
    "liftable",
    "limitFiles",
    "limitProcesses",
    "lines",
    "linkFile",
    "listForm",
    "listSymbols",
    "LLL",
    "lngamma",
    "load",
    "loadPackage",
    "localDictionaries",
    "localize",
    "locate",
    "log",
    "log1p",
    "lookup",
    "lookupCount",
    "LUdecomposition",
    "M2CODE",
    "makeDirectory",
    "makeDocumentTag",
    "makePackageIndex",
    "makeS2",
    "map",
    "markedGB",
    "match",
    "mathML",
    "matrix",
    "max",
    "maxPosition",
    "member",
    "memoize",
    "memoizeClear",
    "memoizeValues",
    "merge",
    "mergePairs",
    "method",
    "methodOptions",
    "methods",
    "midpoint",
    "min",
    "mingens",
    "mingle",
    "minimalBetti",
    "minimalPresentation",
    "minimalPrimes",
    "minimalReduction",
    "minimize",
    "minimizeFilename",
    "minors",
    "minPosition",
    "minPres",
    "minprimes",
    "minus",
    "mkdir",
    "mod",
    "module",
    "modulo",
    "monoid",
    "monomialCurveIdeal",
    "monomialIdeal",
    "monomials",
    "monomialSubideal",
    "moveFile",
    "multidegree",
    "multidoc",
    "multigraded",
    "multiplicity",
    "mutable",
    "mutableIdentity",
    "mutableMatrix",
    "nanosleep",
    "needs",
    "needsPackage",
    "net",
    "netList",
    "newClass",
    "newCoordinateSystem",
    "newNetFile",
    "newPackage",
    "newRing",
    "next",
    "nextkey",
    "nextPrime",
    "NNParser",
    "nonspaceAnalyzer",
    "norm",
    "normalCone",
    "notImplemented",
    "nullhomotopy",
    "nullParser",
    "nullSpace",
    "number",
    "numcols",
    "numColumns",
    "numerator",
    "numeric",
    "numericInterval",
    "numgens",
    "numRows",
    "numrows",
    "odd",
    "oeis",
    "ofClass",
    "on",
    "openDatabase",
    "openDatabaseOut",
    "openFiles",
    "openIn",
    "openInOut",
    "openListener",
    "openOut",
    "openOutAppend",
    "optionalSignParser",
    "options",
    "optP",
    "orP",
    "override",
    "pack",
    "package",
    "packageTemplate",
    "pad",
    "pager",
    "pairs",
    "parent",
    "part",
    "partition",
    "partitions",
    "parts",
    "pdim",
    "peek",
    "permanents",
    "permutations",
    "pfaffians",
    "pivots",
    "plus",
    "poincare",
    "poincareN",
    "polarize",
    "poly",
    "position",
    "positions",
    "power",
    "powermod",
    "precision",
    "preimage",
    "prepend",
    "presentation",
    "pretty",
    "primaryComponent",
    "primaryDecomposition",
    "print",
    "printerr",
    "printString",
    "processID",
    "product",
    "profile",
    "Proj",
    "projectiveHilbertPolynomial",
    "promote",
    "protect",
    "prune",
    "pseudocode",
    "pseudoRemainder",
    "pushForward",
    "QQParser",
    "QRDecomposition",
    "quotient",
    "quotientRemainder",
    "radical",
    "radicalContainment",
    "random",
    "randomKRationalPoint",
    "randomMutableMatrix",
    "rank",
    "read",
    "readDirectory",
    "readlink",
    "readPackage",
    "realPart",
    "realpath",
    "recursionDepth",
    "reducedRowEchelonForm",
    "reduceHilbert",
    "reductionNumber",
    "reesAlgebra",
    "reesAlgebraIdeal",
    "reesIdeal",
    "regex",
    "regexQuote",
    "registerFinalizer",
    "regSeqInIdeal",
    "regularity",
    "regularizedBeta",
    "regularizedGamma",
    "relations",
    "relativizeFilename",
    "remainder",
    "remove",
    "removeDirectory",
    "removeFile",
    "removeLowestDimension",
    "reorganize",
    "replace",
    "res",
    "reshape",
    "resolution",
    "resultant",
    "reverse",
    "right",
    "ring",
    "ringFromFractions",
    "roots",
    "rotate",
    "round",
    "rowAdd",
    "rowMult",
    "rowPermute",
    "rowRankProfile",
    "rowSwap",
    "rsort",
    "run",
    "runHooks",
    "runLengthEncode",
    "runProgram",
    "same",
    "saturate",
    "scan",
    "scanKeys",
    "scanLines",
    "scanPairs",
    "scanValues",
    "schedule",
    "schreyerOrder",
    "Schubert",
    "searchPath",
    "sec",
    "sech",
    "seeParsing",
    "select",
    "selectInSubring",
    "selectVariables",
    "separate",
    "separateRegexp",
    "sequence",
    "serialNumber",
    "set",
    "setEcho",
    "setGroupID",
    "setIOExclusive",
    "setIOSynchronized",
    "setIOUnSynchronized",
    "setRandomSeed",
    "setup",
    "setupEmacs",
    "sheaf",
    "sheafHom",
    "show",
    "showHtml",
    "showTex",
    "simpleDocFrob",
    "sin",
    "singularLocus",
    "sinh",
    "size",
    "size2",
    "sleep",
    "smithNormalForm",
    "solve",
    "someTerms",
    "sort",
    "sortColumns",
    "source",
    "span",
    "Spec",
    "specialFiber",
    "specialFiberIdeal",
    "splice",
    "splitWWW",
    "sqrt",
    "stack",
    "stacksProject",
    "standardForm",
    "standardPairs",
    "stashValue",
    "status",
    "style",
    "sub",
    "sublists",
    "submatrix",
    "submatrixByDegrees",
    "subquotient",
    "subsets",
    "substitute",
    "substring",
    "subtable",
    "sum",
    "super",
    "support",
    "SVD",
    "switch",
    "sylvesterMatrix",
    "symbolBody",
    "symlinkDirectory",
    "symlinkFile",
    "symmetricAlgebra",
    "symmetricAlgebraIdeal",
    "symmetricKernel",
    "symmetricPower",
    "synonym",
    "SYNOPSIS",
    "syz",
    "syzygyScheme",
    "table",
    "take",
    "tally",
    "tan",
    "tangentCone",
    "tangentSheaf",
    "tanh",
    "target",
    "taskResult",
    "temporaryFileName",
    "tensor",
    "tensorAssociativity",
    "terminalParser",
    "terms",
    "TEST",
    "testHunekeQuestion",
    "tests",
    "tex",
    "texMath",
    "times",
    "toAbsolutePath",
    "toCC",
    "toDividedPowers",
    "toDual",
    "toExternalString",
    "toField",
    "toList",
    "toLower",
    "top",
    "topCoefficients",
    "topComponents",
    "toRR",
    "toRRi",
    "toSequence",
    "toString",
    "toUpper",
    "trace",
    "transpose",
    "trim",
    "truncate",
    "truncateOutput",
    "tutorial",
    "ultimate",
    "unbag",
    "uncurry",
    "undocumented",
    "uniform",
    "uninstallAllPackages",
    "uninstallPackage",
    "unique",
    "uniquePermutations",
    "unsequence",
    "unstack",
    "urlEncode",
    "use",
    "userSymbols",
    "utf8",
    "utf8check",
    "utf8substring",
    "validate",
    "value",
    "values",
    "variety",
    "vars",
    "vector",
    "versalEmbedding",
    "wait",
    "wedgeProduct",
    "weightRange",
    "whichGm",
    "width",
    "wikipedia",
    "wrap",
    "youngest",
    "zero",
    "zeta",
    "ZZParser"
    )

M2CONSTANTS = (
    "AbstractToricVarieties",
    "Acknowledgement",
    "AdditionalPaths",
    "AdjointIdeal",
    "AfterEval",
    "AfterNoPrint",
    "AfterPrint",
    "AInfinity",
    "AlgebraicSplines",
    "Algorithm",
    "Alignment",
    "AllCodimensions",
    "allowableThreads",
    "AnalyzeSheafOnP1",
    "applicationDirectorySuffix",
    "argument",
    "Ascending",
    "AssociativeAlgebras",
    "Authors",
    "AuxiliaryFiles",
    "backtrace",
    "Bareiss",
    "BaseFunction",
    "baseRings",
    "BaseRow",
    "BasisElementLimit",
    "Bayer",
    "BeforePrint",
    "BeginningMacaulay2",
    "Benchmark",
    "BernsteinSato",
    "Bertini",
    "BettiCharacters",
    "BGG",
    "BIBasis",
    "Binary",
    "Binomial",
    "BinomialEdgeIdeals",
    "Binomials",
    "BKZ",
    "blockMatrixForm",
    "Body",
    "BoijSoederberg",
    "Book3264Examples",
    "BooleanGB",
    "Boxes",
    "Browse",
    "Bruns",
    "cache",
    "CacheExampleOutput",
    "CallLimit",
    "CannedExample",
    "CatalanConstant",
    "Caveat",
    "CellularResolutions",
    "Center",
    "Certification",
    "ChainComplexExtras",
    "ChainComplexOperations",
    "ChangeMatrix",
    "CharacteristicClasses",
    "CheckDocumentation",
    "Chordal",
    "Classic",
    "clearAll",
    "clearOutput",
    "close",
    "closeIn",
    "closeOut",
    "ClosestFit",
    "Code",
    "CodimensionLimit",
    "CodingTheory",
    "CoefficientRing",
    "Cofactor",
    "CohenEngine",
    "CohenTopLevel",
    "CohomCalg",
    "CoincidentRootLoci",
    "commandLine",
    "compactMatrixForm",
    "Complement",
    "CompleteIntersection",
    "CompleteIntersectionResolutions",
    "Complexes",
    "ConductorElement",
    "Configuration",
    "ConformalBlocks",
    "Consequences",
    "Constants",
    "Contributors",
    "ConvexInterface",
    "ConwayPolynomials",
    "copyright",
    "Core",
    "CorrespondenceScrolls",
    "CotangentSchubert",
    "Cremona",
    "currentFileDirectory",
    "currentFileName",
    "currentLayout",
    "currentPackage",
    "Cyclotomic",
    "Date",
    "dd",
    "DebuggingMode",
    "debuggingMode",
    "debugLevel",
    "DecomposableSparseSystems",
    "Decompose",
    "Default",
    "defaultPrecision",
    "Degree",
    "DegreeGroup",
    "DegreeLift",
    "DegreeLimit",
    "DegreeMap",
    "DegreeOrder",
    "DegreeRank",
    "Degrees",
    "Dense",
    "Density",
    "Depth",
    "Descending",
    "Description",
    "DeterminantalRepresentations",
    "DGAlgebras",
    "dictionaryPath",
    "DiffAlg",
    "Dispatch",
    "DivideConquer",
    "DividedPowers",
    "Divisor",
    "Dmodules",
    "docExample",
    "docTemplate",
    "Down",
    "Dynamic",
    "EagonResolution",
    "EdgeIdeals",
    "edit",
    "EigenSolver",
    "EisenbudHunekeVasconcelos",
    "Elimination",
    "EliminationMatrices",
    "EllipticCurves",
    "EllipticIntegrals",
    "Email",
    "end",
    "endl",
    "Engine",
    "engineDebugLevel",
    "EngineTests",
    "EnumerationCurves",
    "environment",
    "EquivariantGB",
    "errorDepth",
    "EulerConstant",
    "Example",
    "ExampleFiles",
    "ExampleSystems",
    "Exclude",
    "exit",
    "Ext",
    "ExteriorIdeals",
    "ExteriorModules",
    "false",
    "FastMinors",
    "FastNonminimal",
    "FGLM",
    "fileDictionaries",
    "fileExitHooks",
    "FileName",
    "FindOne",
    "FiniteFittingIdeals",
    "First",
    "FirstPackage",
    "FlatMonoid",
    "Flexible",
    "flush",
    "FollowLinks",
    "ForeignFunctions",
    "FormalGroupLaws",
    "Format",
    "FourierMotzkin",
    "FourTiTwo",
    "fpLLL",
    "FrobeniusThresholds",
    "FunctionFieldDesingularization",
    "GBDegrees",
    "gbTrace",
    "GenerateAssertions",
    "Generic",
    "GenericInitialIdeal",
    "GeometricDecomposability",
    "gfanInterface",
    "Givens",
    "GKMVarieties",
    "GLex",
    "Global",
    "GlobalAssignHook",
    "globalAssignmentHooks",
    "GlobalHookStore",
    "GlobalReleaseHook",
    "Gorenstein",
    "GradedLieAlgebras",
    "GraphicalModels",
    "GraphicalModelsMLE",
    "Graphics",
    "Graphs",
    "GRevLex",
    "GroebnerStrata",
    "GroebnerWalk",
    "GroupLex",
    "GroupRevLex",
    "GTZ",
    "Hadamard",
    "handleInterrupts",
    "HardDegreeLimit",
    "Heading",
    "Headline",
    "Heft",
    "Height",
    "help",
    "Hermite",
    "Hermitian",
    "HH",
    "hh",
    "HigherCIOperators",
    "HighestWeights",
    "Hilbert",
    "HodgeIntegrals",
    "HolonomicSystems",
    "homeDirectory",
    "HomePage",
    "Homogeneous",
    "Homogeneous2",
    "HomotopyLieAlgebra",
    "HorizontalSpace",
    "HyperplaneArrangements",
    "id",
    "IgnoreExampleErrors",
    "ii",
    "incomparable",
    "Increment",
    "indeterminate",
    "Index",
    "indexComponents",
    "infinity",
    "InfoDirSection",
    "infoHelp",
    "Inhomogeneous",
    "Inputs",
    "InstallPrefix",
    "IntegralClosure",
    "interpreterDepth",
    "Intersection",
    "InvariantRing",
    "InverseMethod",
    "Inverses",
    "InverseSystems",
    "Invertible",
    "InvolutiveBases",
    "Isomorphism",
    "Item",
    "Iterate",
    "Jacobian",
    "Jets",
    "Join",
    "JSON",
    "Jupyter",
    "K3Carpets",
    "K3Surfaces",
    "Keep",
    "KeepFiles",
    "KeepZeroes",
    "Key",
    "Keywords",
    "Kronecker",
    "KustinMiller",
    "lastMatch",
    "LatticePolytopes",
    "Layout",
    "Left",
    "LengthLimit",
    "Lex",
    "LexIdeals",
    "Licenses",
    "LieTypes",
    "Limit",
    "Linear",
    "LinearAlgebra",
    "LinearTruncations",
    "lineNumber",
    "listLocalSymbols",
    "listUserSymbols",
    "LLLBases",
    "loadDepth",
    "LoadDocumentation",
    "loadedFiles",
    "loadedPackages",
    "Local",
    "LocalRings",
    "LongPolynomial",
    "M0nbar",
    "Macaulay2Doc",
    "MakeDocumentation",
    "MakeHTML",
    "MakeInfo",
    "MakeLinks",
    "MakePDF",
    "MapleInterface",
    "Markov",
    "MatchingFields",
    "Matroids",
    "maxAllowableThreads",
    "maxExponent",
    "MaximalRank",
    "MaxReductionCount",
    "MCMApproximations",
    "MergeTeX",
    "minExponent",
    "MinimalGenerators",
    "MinimalMatrix",
    "minimalPresentationMap",
    "minimalPresentationMapInv",
    "MinimalPrimes",
    "Minimize",
    "MinimumVersion",
    "Miura",
    "MixedMultiplicity",
    "ModuleDeformations",
    "MonodromySolver",
    "Monomial",
    "MonomialAlgebras",
    "MonomialIntegerPrograms",
    "MonomialOrbits",
    "MonomialOrder",
    "Monomials",
    "MonomialSize",
    "MultiGradedRationalMap",
    "MultiplicitySequence",
    "MultiplierIdeals",
    "MultiplierIdealsDim2",
    "MultiprojectiveVarieties",
    "NAGtypes",
    "Name",
    "Nauty",
    "NautyGraphs",
    "NCAlgebra",
    "NCLex",
    "NewFromMethod",
    "newline",
    "NewMethod",
    "NewOfFromMethod",
    "NewOfMethod",
    "nil",
    "Node",
    "NoetherianOperators",
    "NoetherNormalization",
    "NonminimalComplexes",
    "NoPrint",
    "Normaliz",
    "NormalToricVarieties",
    "notify",
    "NTL",
    "null",
    "nullaryMethods",
    "NumericalAlgebraicGeometry",
    "NumericalCertification",
    "NumericalImplicitization",
    "NumericalLinearAlgebra",
    "NumericalSchubertCalculus",
    "NumericSolutions",
    "OldPolyhedra",
    "OldToricVectorBundles",
    "OnlineLookup",
    "OO",
    "oo",
    "ooo",
    "oooo",
    "OpenMath",
    "operatorAttributes",
    "OptionalComponentsPresent",
    "Options",
    "Order",
    "order",
    "OutputDictionary",
    "Outputs",
    "PackageCitations",
    "PackageDictionary",
    "PackageExports",
    "PackageImports",
    "PackageTemplate",
    "PairLimit",
    "PairsRemaining",
    "Parametrization",
    "Parsing",
    "path",
    "PencilsOfQuadrics",
    "Permanents",
    "PHCpack",
    "PhylogeneticTrees",
    "pi",
    "PieriMaps",
    "PlaneCurveSingularities",
    "Points",
    "Polyhedra",
    "Polymake",
    "PolyominoIdeals",
    "Posets",
    "Position",
    "PositivityToricBundles",
    "POSIX",
    "Postfix",
    "Pre",
    "Precision",
    "Prefix",
    "prefixDirectory",
    "prefixPath",
    "PrimaryDecomposition",
    "PrimaryTag",
    "PrimitiveElement",
    "Print",
    "printingAccuracy",
    "printingLeadLimit",
    "printingPrecision",
    "printingSeparator",
    "printingTimeLimit",
    "printingTrailLimit",
    "printWidth",
    "Probability",
    "profileSummary",
    "programPaths",
    "Projective",
    "Prune",
    "PruneComplex",
    "pruningMap",
    "PseudomonomialPrimaryDecomposition",
    "Pullback",
    "PushForward",
    "Python",
    "QthPower",
    "Quasidegrees",
    "QuaternaryQuartics",
    "QuillenSuslin",
    "quit",
    "Quotient",
    "Radical",
    "RadicalCodim1",
    "RaiseError",
    "RandomCanonicalCurves",
    "RandomComplexes",
    "RandomCurves",
    "RandomCurvesOverVerySmallFiniteFields",
    "RandomGenus14Curves",
    "RandomIdeals",
    "RandomMonomialIdeals",
    "RandomObjects",
    "RandomPlaneCurves",
    "RandomPoints",
    "RandomSpaceCurves",
    "Range",
    "RationalMaps",
    "RationalPoints",
    "RationalPoints2",
    "ReactionNetworks",
    "RealFP",
    "RealQP",
    "RealQP1",
    "RealRoots",
    "RealRR",
    "RealXD",
    "recursionLimit",
    "Reduce",
    "ReesAlgebra",
    "References",
    "ReflexivePolytopesDB",
    "Regularity",
    "RelativeCanonicalResolution",
    "Reload",
    "RemakeAllDocumentation",
    "RerunExamples",
    "ResidualIntersections",
    "ResLengthThree",
    "ResolutionsOfStanleyReisnerRings",
    "restart",
    "Result",
    "Resultants",
    "returnCode",
    "Reverse",
    "RevLex",
    "Right",
    "rootPath",
    "rootURI",
    "RunDirectory",
    "RunExamples",
    "RunExternalM2",
    "SagbiGbDetection",
    "Saturation",
    "Schubert2",
    "SchurComplexes",
    "SchurFunctors",
    "SchurRings",
    "scriptCommandLine",
    "SCSCP",
    "SectionRing",
    "SeeAlso",
    "SegreClasses",
    "SemidefiniteProgramming",
    "Seminormalization",
    "SeparateExec",
    "Serialization",
    "sheafExt",
    "ShimoyamaYokoyama",
    "showClassStructure",
    "showStructure",
    "showUserStructure",
    "SimpleDoc",
    "SimplicialComplexes",
    "SimplicialDecomposability",
    "SimplicialPosets",
    "SimplifyFractions",
    "SizeLimit",
    "SkewCommutative",
    "SlackIdeals",
    "SLnEquivariantMatrices",
    "SLPexpressions",
    "Sort",
    "SortStrategy",
    "SourceCode",
    "SourceRing",
    "SpaceCurves",
    "SparseResultants",
    "SpechtModule",
    "SpecialFanoFourfolds",
    "SpectralSequences",
    "SRdeformations",
    "Standard",
    "StartWithOneMinor",
    "StatePolytope",
    "StatGraphs",
    "stderr",
    "stdio",
    "StopBeforeComputation",
    "stopIfError",
    "StopIteration",
    "StopWithMinimalGenerators",
    "Strategy",
    "Strict",
    "StronglyStableIdeals",
    "Style",
    "SubalgebraBases",
    "Subnodes",
    "SubringLimit",
    "subscript",
    "Sugarless",
    "SumsOfSquares",
    "SuperLinearAlgebra",
    "superscript",
    "SVDComplexes",
    "SwitchingFields",
    "SymbolicPowers",
    "SymmetricPolynomials",
    "Synopsis",
    "Syzygies",
    "SyzygyLimit",
    "SyzygyMatrix",
    "SyzygyRows",
    "TangentCone",
    "TateOnProducts",
    "TensorComplexes",
    "Test",
    "testExample",
    "TestIdeals",
    "TeXmacs",
    "Text",
    "ThinSincereQuivers",
    "ThreadedGB",
    "Threshold",
    "Topcom",
    "topLevelMode",
    "Tor",
    "TorAlgebra",
    "Toric",
    "ToricInvariants",
    "ToricTopology",
    "ToricVectorBundles",
    "Torsion",
    "TotalPairs",
    "Tree",
    "TriangularSets",
    "Triangulations",
    "Tries",
    "Trim",
    "Triplets",
    "Tropical",
    "true",
    "Truncate",
    "Truncations",
    "TSpreadIdeals",
    "TypicalValue",
    "typicalValues",
    "Undo",
    "Unique",
    "Units",
    "Unmixed",
    "Up",
    "UpdateOnly",
    "UpperTriangular",
    "Usage",
    "UseCachedExampleOutput",
    "UseHilbertFunction",
    "UserMode",
    "UseSyzygies",
    "Variable",
    "VariableBaseName",
    "Variables",
    "Vasconcelos",
    "VectorFields",
    "VectorGraphics",
    "Verbose",
    "Verbosity",
    "Verify",
    "VersalDeformations",
    "Version",
    "version",
    "VerticalSpace",
    "viewHelp",
    "VirtualResolutions",
    "Visualize",
    "WebApp",
    "Weights",
    "WeylAlgebra",
    "WeylGroups",
    "WhitneyStratifications",
    "Wrap",
    "XML"
    )

class Macaulay2Lexer(RegexLexer):
    """Lexer for Macaulay2, a software system for research in algebraic geometry."""

    name = 'Macaulay2'
    url = 'https://faculty.math.illinois.edu/Macaulay2/'
    aliases = ['macaulay2']
    filenames = ['*.m2']

    tokens = {
        'root': [
            (r'--.*$', Comment.Single),
            (r'-\*', Comment.Multiline, 'block comment'),
            (r'"', String, 'quote string'),
            (r'///', String, 'slash string'),
            (words(M2KEYWORDS, prefix=r'\b', suffix=r'\b'), Keyword),
            (words(M2DATATYPES, prefix=r'\b', suffix=r'\b'), Name.Builtin),
            (words(M2FUNCTIONS, prefix=r'\b', suffix=r'\b'), Name.Function),
            (words(M2CONSTANTS, prefix=r'\b', suffix=r'\b'), Name.Constant),
            (r'\s+', Text.Whitespace),
            (r'.', Text)
        ],
        'block comment' : [
            (r'[^*-]+', Comment.Multiline),
            (r'\*-', Comment.Multiline, '#pop'),
            (r'[*-]', Comment.Multiline)
        ],
        'quote string' : [
            (r'[^\\"]+', String),
            (r'"', String, '#pop'),
            (r'\\"?', String),
        ],
        'slash string' : [
            (r'[^/]+', String),
            (r'(//)+(?!/)', String),
            (r'/(//)+(?!/)', String, '#pop'),
            (r'/', String)
        ]
    }



import base64
import os
import smtplib
from datetime import datetime
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import httplib2
from apiclient import discovery
from oauth2client import client, tools, file

from ..utils.reporter_utils import format_error


class EmailSender:
    def __init__(self, email=None, password=None, server_smtp=None, server_port=None, gmail_token=None, debug=True):
        """
        General init

        :param email: email of user, from which emails will be sent, string
        :param password: password of user, from which emails will be sent, string
        :param server_smtp: full smtp address (endpoint) of mail server, string
        :param server_port: mail server port, integer
        :gmail_token: gmail OAuth secret file (expected json)
        :param debug: debug output is enabled, may be True or False, optional
        """
        if debug:
            print("Email Sender init")
        self.__debug = debug
        self.__method = None
        if email is not None and password is not None and server_smtp is not None and server_port is not None:
            self.__method = 'regular'
        elif gmail_token and email:
            gmail_token = f'{os.getcwd()}/{gmail_token}' if not os.path.exists(gmail_token) else gmail_token
            if os.path.exists(gmail_token):
                self.__method = 'gmail'
                self.__gmail_scopes = 'https://www.googleapis.com/auth/gmail.send'
                self.__gmail_app_name = 'Gmail API Python Send Email'
        if not self.__method:
            raise ValueError("No email credentials are provided, aborted!")
        self.__email = email
        self.__password = password
        self.__server_smtp = server_smtp
        self.__server_port = server_port
        self.__gmail_token = gmail_token

    def send_message(self, files=None, captions=None, image_width="400px", title=None, timestamp=None, recipients=None,
                     method=None, custom_message=None,
                     custom_folder=os.path.join(os.path.expanduser('~'), '.credentials'), debug=None):
        """
        Send email to recipients with report (with attached images)

        :param files: list of filenames (maybe with path) with charts to attach to report, list of strings, required
        :param captions: captions for charts, length should be equal to count of files, list of strings, optional
        :param image_width: default image width, string, optional
        :param title: title of report, string, optional
        :param timestamp: non-default timestamp, string, optional, will be used only when title is not provided
        :param recipients: list of recipient emails, list of strings, optional
        :param method: method which will be used for sending
        :param custom_message: custom message, prepared by user at his own, by default its payload with TR state report
        :param custom_folder: custom home folder for gmail credentials storage, by default is ~/.credentials
        :param debug: debug output is enabled, may be True or False, optional
        :return: none
        """
        # Check params
        if not method:
            method = self.__method
        if not isinstance(files, list) and not custom_message:
            raise ValueError("No file list for report provided, aborted!")
        if isinstance(recipients, str) and not custom_message:
            recipients = [recipients]
        elif not isinstance(recipients, list) and not custom_message:
            raise ValueError("Wrong list of recipients is provided, aborted!")
        debug = debug if debug is not None else self.__debug
        if not isinstance(captions, list) or custom_message:
            if debug:
                print("Caption list is empty, no legend will be displayed")
            captions = None
        elif len(captions) != len(files):
            if debug:
                print(f"Caption and file lists are not the same length {len(captions)} != {len(files)} thus "
                      f"no legend will be displayed")
            captions = None
        timestamp = timestamp if timestamp else datetime.now().strftime("%Y-%m-%d")
        title = title if title else f"Test development & automation coverage report for {timestamp}"

        # Connect and send message
        if not custom_message:
            message = self.__prepare_payload(files=files, captions=captions, image_width=image_width, title=title,
                                             recipients=recipients, method=method)
        else:
            if debug:
                print("Ignoring payload preparations, assuming user custom message is right")
            message = custom_message
        if method == 'regular':
            connection = self.__connect_to_server()
            self.__send_to_server(connection=connection, recipients=recipients, message=message)
            self.__disconnect_from_server(connection=connection)
        elif method == 'gmail':
            self.__gmail_send_message(message=message, custom_folder=custom_folder)
        if debug:
            print("Email sent!")

    def __connect_to_server(self):
        """
        Connects to mail server

        :return: connection handle ( smtplib.SMTP )
        """
        if self.__debug:
            print(f'Connecting to custom mail server {self.__server_smtp}:{self.__server_port} using {self.__email}')
        try:
            connection = smtplib.SMTP(self.__server_smtp, self.__server_port)
            connection.ehlo()
            connection.starttls()
            connection.login(self.__email, self.__password)
        except Exception as e:
            raise ValueError(f"Can't login to mail!\nError{format_error(e)}")
        return connection

    def __send_to_server(self, connection, recipients, message):
        """
        Send data to server to mail server

        :param connection: connection handle ( smtplib.SMTP )
        :param recipients: list of recipient emails, list of strings, optional
        :param message: formatted multipart message
        :return: none
        """
        if self.__debug:
            print(f'Sending mail from {self.__email} to {recipients}')
        try:
            connection.sendmail(from_addr=self.__email, to_addrs=recipients, msg=message.as_string())
        except Exception as e:
            raise ValueError(f"Can't send mail!\nError{format_error(e)}")

    def __disconnect_from_server(self, connection):
        """
        Connects to mail server

        :param connection: connection handle ( smtplib.SMTP )
        :return: none
        """
        if self.__debug:
            print(f'Disconnecting from custom server')
        try:
            connection.quit()
        except Exception as e:
            raise ValueError(f"Can't close connection!\nError{format_error(e)}")

    def __prepare_payload(self, files, image_width, title, recipients, captions=None, method=None):
        """
        Prepare payload method (mail content)

        :param files: list of filenames (maybe with path) with charts to attach to report, list of strings, required
        :param captions: captions for charts, length should be equal to count of files, list of strings, optional
        :param image_width: default image width, string
        :param title: title of report, string
        :param recipients: list of recipient emails, list of strings, optional
        :param method: specify which method is used to set proper MIMEMultipart type ('gmail' or not)
        :return: formatted multipart message
        """
        message = MIMEMultipart("alternative") if method != 'gmail' else MIMEMultipart()
        message["Subject"] = title
        message["From"] = self.__email
        message["To"] = ", ".join(recipients)
        html = '<!DOCTYPE html><html lang="en"><head><meta charset="UTF-8">' \
               f'<title>{title}</title></head>' \
               f'<body><div align="center"><h3>{title}' \
               '</h3></div><table border="0px" width="100%"><tbody><td>'
        for j, val in enumerate(files):
            with open('{}'.format(val), "rb") as attachment:
                mime_image = MIMEImage(attachment.read())

            # Define the image's ID with counter as you will reference it.
            mime_image.add_header('Content-ID', f'<image_id_{j}>')
            mime_image.add_header('Content-Disposition', f'attachment; filename= {val}')
            message.attach(mime_image)
            # add to body
            if captions:
                html = f'{html}<tr><div align="center"><b>{captions[j]}</b></div></tr>'
            html = f'{html}<tr><div align="center"><img src="cid:image_id_{j}" ' \
                   f'width="{image_width}" height="auto">></div></tr>'
        html = f'{html}</td></tbody></table></body></html>'
        message.attach(MIMEText(html, "html"))
        return message

    def __gmail_get_credential_path(self, custom_folder=os.path.join(os.path.expanduser('~'), '.credentials')):
        """
        Service function target Google OAuth credentials path to storage

        :param custom_folder: custom home folder for gmail credentials storage, by default is ~/.credentials
        :return: credentials file path (string)
        """
        if self.__debug:
            print(f"Checking GMail credentials path at {custom_folder}")
        try:
            if self.__debug:
                print(f"No credential directory found, creating new one here: {custom_folder}")
            os.makedirs(custom_folder, exist_ok=True)
        except OSError as e:
            if self.__debug:
                print(f"Original Error{format_error(e)}")
        credential_path = os.path.join(custom_folder, 'gmail-python-email-send.json')
        return credential_path

    def __gmail_get_credentials(self, custom_folder=os.path.join(os.path.expanduser('~'), '.credentials')):
        """
        Service function to get and convert Google OAuth credential from client_id and client_secret

        :param custom_folder: custom home folder for gmail credentials storage, by default is ~/.credentials
        :return: credentials
        """
        credential_path = self.__gmail_get_credential_path(custom_folder=custom_folder)
        if self.__debug:
            print(f"Obtaining GMail credentials from {credential_path}")
        try:
            store = file.Storage(credential_path)
        except Exception as e:
            raise ValueError(f"Couldn't open storage\nError{format_error(e)}")
        try:
            credentials = store.get()
        except Exception as e:
            raise ValueError(f"Obtaining of credentials unexpectedly failed\nError{format_error(e)}")
        if not credentials or credentials.invalid:
            try:
                flow = client.flow_from_clientsecrets(self.__gmail_token, self.__gmail_scopes)
            except Exception as e:
                raise ValueError(f"Couldn't obtain new client secrets from Google OAuth\nError{format_error(e)}")
            flow.user_agent = self.__gmail_app_name
            try:
                credentials = tools.run_flow(flow, store)
            except Exception as e:
                raise ValueError(f"Couldn't obtain new credential from Google OAuth\nError{format_error(e)}")
            if self.__debug:
                print('Credentials stored to ' + credential_path)
        return credentials

    def __gmail_send_message(self, message, custom_folder=os.path.join(os.path.expanduser('~'), '.credentials')):
        """
        Send Email via GMail

        :param message: message in MIME type format
        :param custom_folder: custom home folder for gmail credentials storage, by default is ~/.credentials
        :return: none
        """
        if self.__debug:
            print("Sending message using GMail")
        credentials = self.__gmail_get_credentials(custom_folder=custom_folder)
        try:
            http = credentials.authorize(httplib2.Http())
        except Exception as e:
            raise ValueError(f"Can't authorize via Google OAuth\nError{format_error(e)}")
        try:
            service = discovery.build('gmail', 'v1', http=http)
        except Exception as e:
            raise ValueError(f"Can't build service for Google OAuth\nError{format_error(e)}")
        try:
            raw = base64.urlsafe_b64encode(message.as_bytes()).decode()
        except Exception as e:
            raise ValueError(f"Can't convert payload to base64\nError{format_error(e)}")
        self.__gmail_send_message_internal(service, self.__email, {'raw': raw})

    def __gmail_send_message_internal(self, service, user_id, message):
        """
        Low-level gmail sent function to send email via GMail API service

        :param service: service API
        :param user_id: user id, the same as "from" email field
        :param message: formatted in base64 type encoded raw message
        :return: message
        """
        try:
            message = (service.users().messages().send(userId=user_id, body=message).execute())
            if self.__debug:
                print(f'Message sent with Id: "{message["id"]}"!')
            return message
        except Exception as e:
            raise ValueError(f"Can't send mail via GMail!\nError{format_error(e)}")



"""
    pygments.lexers.nix
    ~~~~~~~~~~~~~~~~~~~

    Lexers for the NixOS Nix language.

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re

from pygments.lexer import RegexLexer, include
from pygments.token import Text, Comment, Operator, Keyword, Name, String, \
    Number, Punctuation, Literal

__all__ = ['NixLexer']


class NixLexer(RegexLexer):
    """
    For the Nix language.

    .. versionadded:: 2.0
    """

    name = 'Nix'
    url = 'http://nixos.org/nix/'
    aliases = ['nixos', 'nix']
    filenames = ['*.nix']
    mimetypes = ['text/x-nix']

    keywords = ['rec', 'with', 'let', 'in', 'inherit', 'assert', 'if',
                'else', 'then', '...']
    builtins = ['import', 'abort', 'baseNameOf', 'dirOf', 'isNull', 'builtins',
                'map', 'removeAttrs', 'throw', 'toString', 'derivation']
    operators = ['++', '+', '?', '.', '!', '//', '==',
                 '!=', '&&', '||', '->', '=']

    punctuations = ["(", ")", "[", "]", ";", "{", "}", ":", ",", "@"]

    tokens = {
        'root': [
            # comments starting with #
            (r'#.*$', Comment.Single),

            # multiline comments
            (r'/\*', Comment.Multiline, 'comment'),

            # whitespace
            (r'\s+', Text),

            # keywords
            ('(%s)' % '|'.join(re.escape(entry) + '\\b' for entry in keywords), Keyword),

            # highlight the builtins
            ('(%s)' % '|'.join(re.escape(entry) + '\\b' for entry in builtins),
             Name.Builtin),

            (r'\b(true|false|null)\b', Name.Constant),

            # operators
            ('(%s)' % '|'.join(re.escape(entry) for entry in operators),
             Operator),

            # word operators
            (r'\b(or|and)\b', Operator.Word),

            # punctuations
            ('(%s)' % '|'.join(re.escape(entry) for entry in punctuations), Punctuation),

            # integers
            (r'[0-9]+', Number.Integer),

            # strings
            (r'"', String.Double, 'doublequote'),
            (r"''", String.Single, 'singlequote'),

            # paths
            (r'[\w.+-]*(\/[\w.+-]+)+', Literal),
            (r'\<[\w.+-]+(\/[\w.+-]+)*\>', Literal),

            # urls
            (r'[a-zA-Z][a-zA-Z0-9\+\-\.]*\:[\w%/?:@&=+$,\\.!~*\'-]+', Literal),

            # names of variables
            (r'[\w-]+\s*=', String.Symbol),
            (r'[a-zA-Z_][\w\'-]*', Text),

        ],
        'comment': [
            (r'[^/*]+', Comment.Multiline),
            (r'/\*', Comment.Multiline, '#push'),
            (r'\*/', Comment.Multiline, '#pop'),
            (r'[*/]', Comment.Multiline),
        ],
        'singlequote': [
            (r"'''", String.Escape),
            (r"''\$\{", String.Escape),
            (r"''\n", String.Escape),
            (r"''\r", String.Escape),
            (r"''\t", String.Escape),
            (r"''", String.Single, '#pop'),
            (r'\$\{', String.Interpol, 'antiquote'),
            (r"['$]", String.Single),
            (r"[^'$]+", String.Single),
        ],
        'doublequote': [
            (r'\\', String.Escape),
            (r'\\"', String.Escape),
            (r'\\$\{', String.Escape),
            (r'"', String.Double, '#pop'),
            (r'\$\{', String.Interpol, 'antiquote'),
            (r'[^"]', String.Double),
        ],
        'antiquote': [
            (r"\}", String.Interpol, '#pop'),
            # TODO: we should probably escape also here ''${ \${
            (r"\$\{", String.Interpol, '#push'),
            include('root'),
        ],
    }

    def analyse_text(text):
        rv = 0.0
        # TODO: let/in
        if re.search(r'import.+?<[^>]+>', text):
            rv += 0.4
        if re.search(r'mkDerivation\s+(\(|\{|rec)', text):
            rv += 0.4
        if re.search(r'=\s+mkIf\s+', text):
            rv += 0.4
        if re.search(r'\{[a-zA-Z,\s]+\}:', text):
            rv += 0.1
        return rv



"""
    pygments.styles.friendly_grayscale
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    A style based on friendly style.
    The color values of the friendly style have been converted to grayscale
    using the luminosity value calculated by
    http://www.workwithcolor.com/color-converter-01.htm

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, \
     Number, Operator, Generic, Whitespace


class FriendlyGrayscaleStyle(Style):
    """
    A modern grayscale style based on the friendly style.

    .. versionadded:: 2.11
    """

    background_color = "#f0f0f0"

    styles = {
        Whitespace:                "#bbbbbb",
        Comment:                   "italic #959595",
        Comment.Preproc:           "noitalic #575757",
        Comment.Special:           "noitalic bg:#F4F4F4",

        Keyword:                   "bold #575757",
        Keyword.Pseudo:            "nobold",
        Keyword.Type:              "nobold #4F4F4F",

        Operator:                  "#666666",
        Operator.Word:             "bold #575757",

        Name.Builtin:              "#575757",
        Name.Function:             "#3F3F3F",
        Name.Class:                "bold #7E7E7E",
        Name.Namespace:            "bold #7E7E7E",
        Name.Exception:            "#575757",
        Name.Variable:             "#9A9A9A",
        Name.Constant:             "#A5A5A5",
        Name.Label:                "bold #363636",
        Name.Entity:               "bold #848484",
        Name.Attribute:            "#707070",
        Name.Tag:                  "bold #3B3B3B",
        Name.Decorator:            "bold #555555",

        String:                    "#717171",
        String.Doc:                "italic",
        String.Interpol:           "italic #9F9F9F",
        String.Escape:             "bold #717171",
        String.Regex:              "#575757",
        String.Symbol:             "#676767",
        String.Other:              "#7E7E7E",
        Number:                    "#888888",

        Generic.Heading:           "bold #373737",
        Generic.Subheading:        "bold #5A5A5A",
        Generic.Deleted:           "#545454",
        Generic.Inserted:          "#7D7D7D",
        Generic.Error:             "#898989",
        Generic.Emph:              "italic",
        Generic.Strong:            "bold",
        Generic.EmphStrong:        "bold italic",
        Generic.Prompt:            "bold #7E7E7E",
        Generic.Output:            "#888888",
        Generic.Traceback:         "#6D6D6D",

        Error:                     "border:#898989"
    }



"""
    pygments.lexers.sas
    ~~~~~~~~~~~~~~~~~~~

    Lexer for SAS.

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re
from pygments.lexer import RegexLexer, include, words
from pygments.token import Comment, Keyword, Name, Number, String, Text, \
    Other, Generic

__all__ = ['SASLexer']


class SASLexer(RegexLexer):
    """
    For SAS files.

    .. versionadded:: 2.2
    """
    # Syntax from syntax/sas.vim by James Kidd <james.kidd@covance.com>

    name      = 'SAS'
    aliases   = ['sas']
    filenames = ['*.SAS', '*.sas']
    mimetypes = ['text/x-sas', 'text/sas', 'application/x-sas']
    flags     = re.IGNORECASE | re.MULTILINE

    builtins_macros = (
        "bquote", "nrbquote", "cmpres", "qcmpres", "compstor", "datatyp",
        "display", "do", "else", "end", "eval", "global", "goto", "if",
        "index", "input", "keydef", "label", "left", "length", "let",
        "local", "lowcase", "macro", "mend", "nrquote",
        "nrstr", "put", "qleft", "qlowcase", "qscan",
        "qsubstr", "qsysfunc", "qtrim", "quote", "qupcase", "scan",
        "str", "substr", "superq", "syscall", "sysevalf", "sysexec",
        "sysfunc", "sysget", "syslput", "sysprod", "sysrc", "sysrput",
        "then", "to", "trim", "unquote", "until", "upcase", "verify",
        "while", "window"
    )

    builtins_conditionals = (
        "do", "if", "then", "else", "end", "until", "while"
    )

    builtins_statements = (
        "abort", "array", "attrib", "by", "call", "cards", "cards4",
        "catname", "continue", "datalines", "datalines4", "delete", "delim",
        "delimiter", "display", "dm", "drop", "endsas", "error", "file",
        "filename", "footnote", "format", "goto", "in", "infile", "informat",
        "input", "keep", "label", "leave", "length", "libname", "link",
        "list", "lostcard", "merge", "missing", "modify", "options", "output",
        "out", "page", "put", "redirect", "remove", "rename", "replace",
        "retain", "return", "select", "set", "skip", "startsas", "stop",
        "title", "update", "waitsas", "where", "window", "x", "systask"
    )

    builtins_sql = (
        "add", "and", "alter", "as", "cascade", "check", "create",
        "delete", "describe", "distinct", "drop", "foreign", "from",
        "group", "having", "index", "insert", "into", "in", "key", "like",
        "message", "modify", "msgtype", "not", "null", "on", "or",
        "order", "primary", "references", "reset", "restrict", "select",
        "set", "table", "unique", "update", "validate", "view", "where"
    )

    builtins_functions = (
        "abs", "addr", "airy", "arcos", "arsin", "atan", "attrc",
        "attrn", "band", "betainv", "blshift", "bnot", "bor",
        "brshift", "bxor", "byte", "cdf", "ceil", "cexist", "cinv",
        "close", "cnonct", "collate", "compbl", "compound",
        "compress", "cos", "cosh", "css", "curobs", "cv", "daccdb",
        "daccdbsl", "daccsl", "daccsyd", "dacctab", "dairy", "date",
        "datejul", "datepart", "datetime", "day", "dclose", "depdb",
        "depdbsl", "depsl", "depsyd",
        "deptab", "dequote", "dhms", "dif", "digamma",
        "dim", "dinfo", "dnum", "dopen", "doptname", "doptnum",
        "dread", "dropnote", "dsname", "erf", "erfc", "exist", "exp",
        "fappend", "fclose", "fcol", "fdelete", "fetch", "fetchobs",
        "fexist", "fget", "fileexist", "filename", "fileref",
        "finfo", "finv", "fipname", "fipnamel", "fipstate", "floor",
        "fnonct", "fnote", "fopen", "foptname", "foptnum", "fpoint",
        "fpos", "fput", "fread", "frewind", "frlen", "fsep", "fuzz",
        "fwrite", "gaminv", "gamma", "getoption", "getvarc", "getvarn",
        "hbound", "hms", "hosthelp", "hour", "ibessel", "index",
        "indexc", "indexw", "input", "inputc", "inputn", "int",
        "intck", "intnx", "intrr", "irr", "jbessel", "juldate",
        "kurtosis", "lag", "lbound", "left", "length", "lgamma",
        "libname", "libref", "log", "log10", "log2", "logpdf", "logpmf",
        "logsdf", "lowcase", "max", "mdy", "mean", "min", "minute",
        "mod", "month", "mopen", "mort", "n", "netpv", "nmiss",
        "normal", "note", "npv", "open", "ordinal", "pathname",
        "pdf", "peek", "peekc", "pmf", "point", "poisson", "poke",
        "probbeta", "probbnml", "probchi", "probf", "probgam",
        "probhypr", "probit", "probnegb", "probnorm", "probt",
        "put", "putc", "putn", "qtr", "quote", "ranbin", "rancau",
        "ranexp", "rangam", "range", "rank", "rannor", "ranpoi",
        "rantbl", "rantri", "ranuni", "repeat", "resolve", "reverse",
        "rewind", "right", "round", "saving", "scan", "sdf", "second",
        "sign", "sin", "sinh", "skewness", "soundex", "spedis",
        "sqrt", "std", "stderr", "stfips", "stname", "stnamel",
        "substr", "sum", "symget", "sysget", "sysmsg", "sysprod",
        "sysrc", "system", "tan", "tanh", "time", "timepart", "tinv",
        "tnonct", "today", "translate", "tranwrd", "trigamma",
        "trim", "trimn", "trunc", "uniform", "upcase", "uss", "var",
        "varfmt", "varinfmt", "varlabel", "varlen", "varname",
        "varnum", "varray", "varrayx", "vartype", "verify", "vformat",
        "vformatd", "vformatdx", "vformatn", "vformatnx", "vformatw",
        "vformatwx", "vformatx", "vinarray", "vinarrayx", "vinformat",
        "vinformatd", "vinformatdx", "vinformatn", "vinformatnx",
        "vinformatw", "vinformatwx", "vinformatx", "vlabel",
        "vlabelx", "vlength", "vlengthx", "vname", "vnamex", "vtype",
        "vtypex", "weekday", "year", "yyq", "zipfips", "zipname",
        "zipnamel", "zipstate"
    )

    tokens = {
        'root': [
            include('comments'),
            include('proc-data'),
            include('cards-datalines'),
            include('logs'),
            include('general'),
            (r'.', Text),
        ],
        # SAS is multi-line regardless, but * is ended by ;
        'comments': [
            (r'^\s*\*.*?;', Comment),
            (r'/\*.*?\*/', Comment),
            (r'^\s*\*(.|\n)*?;', Comment.Multiline),
            (r'/[*](.|\n)*?[*]/', Comment.Multiline),
        ],
        # Special highlight for proc, data, quit, run
        'proc-data': [
            (r'(^|;)\s*(proc \w+|data|run|quit)[\s;]',
             Keyword.Reserved),
        ],
        # Special highlight cards and datalines
        'cards-datalines': [
            (r'^\s*(datalines|cards)\s*;\s*$', Keyword, 'data'),
        ],
        'data': [
            (r'(.|\n)*^\s*;\s*$', Other, '#pop'),
        ],
        # Special highlight for put NOTE|ERROR|WARNING (order matters)
        'logs': [
            (r'\n?^\s*%?put ', Keyword, 'log-messages'),
        ],
        'log-messages': [
            (r'NOTE(:|-).*', Generic, '#pop'),
            (r'WARNING(:|-).*', Generic.Emph, '#pop'),
            (r'ERROR(:|-).*', Generic.Error, '#pop'),
            include('general'),
        ],
        'general': [
            include('keywords'),
            include('vars-strings'),
            include('special'),
            include('numbers'),
        ],
        # Keywords, statements, functions, macros
        'keywords': [
            (words(builtins_statements,
                   prefix = r'\b',
                   suffix = r'\b'),
             Keyword),
            (words(builtins_sql,
                   prefix = r'\b',
                   suffix = r'\b'),
             Keyword),
            (words(builtins_conditionals,
                   prefix = r'\b',
                   suffix = r'\b'),
             Keyword),
            (words(builtins_macros,
                   prefix = r'%',
                   suffix = r'\b'),
             Name.Builtin),
            (words(builtins_functions,
                   prefix = r'\b',
                   suffix = r'\('),
             Name.Builtin),
        ],
        # Strings and user-defined variables and macros (order matters)
        'vars-strings': [
            (r'&[a-z_]\w{0,31}\.?', Name.Variable),
            (r'%[a-z_]\w{0,31}', Name.Function),
            (r'\'', String, 'string_squote'),
            (r'"', String, 'string_dquote'),
        ],
        'string_squote': [
            ('\'', String, '#pop'),
            (r'\\\\|\\"|\\\n', String.Escape),
            # AFAIK, macro variables are not evaluated in single quotes
            # (r'&', Name.Variable, 'validvar'),
            (r'[^$\'\\]+', String),
            (r'[$\'\\]', String),
        ],
        'string_dquote': [
            (r'"', String, '#pop'),
            (r'\\\\|\\"|\\\n', String.Escape),
            (r'&', Name.Variable, 'validvar'),
            (r'[^$&"\\]+', String),
            (r'[$"\\]', String),
        ],
        'validvar': [
            (r'[a-z_]\w{0,31}\.?', Name.Variable, '#pop'),
        ],
        # SAS numbers and special variables
        'numbers': [
            (r'\b[+-]?([0-9]+(\.[0-9]+)?|\.[0-9]+|\.)(E[+-]?[0-9]+)?i?\b',
             Number),
        ],
        'special': [
            (r'(null|missing|_all_|_automatic_|_character_|_n_|'
             r'_infile_|_name_|_null_|_numeric_|_user_|_webout_)',
             Keyword.Constant),
        ],
        # 'operators': [
        #     (r'(-|=|<=|>=|<|>|<>|&|!=|'
        #      r'\||\*|\+|\^|/|!|~|~=)', Operator)
        # ],
    }



"""
    pygments.formatter
    ~~~~~~~~~~~~~~~~~~

    Base formatter class.

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import codecs

from pygments.util import get_bool_opt
from pygments.styles import get_style_by_name

__all__ = ['Formatter']


def _lookup_style(style):
    if isinstance(style, str):
        return get_style_by_name(style)
    return style


class Formatter:
    """
    Converts a token stream to text.

    Formatters should have attributes to help selecting them. These
    are similar to the corresponding :class:`~pygments.lexer.Lexer`
    attributes.

    .. autoattribute:: name
       :no-value:

    .. autoattribute:: aliases
       :no-value:

    .. autoattribute:: filenames
       :no-value:

    You can pass options as keyword arguments to the constructor.
    All formatters accept these basic options:

    ``style``
        The style to use, can be a string or a Style subclass
        (default: "default"). Not used by e.g. the
        TerminalFormatter.
    ``full``
        Tells the formatter to output a "full" document, i.e.
        a complete self-contained document. This doesn't have
        any effect for some formatters (default: false).
    ``title``
        If ``full`` is true, the title that should be used to
        caption the document (default: '').
    ``encoding``
        If given, must be an encoding name. This will be used to
        convert the Unicode token strings to byte strings in the
        output. If it is "" or None, Unicode strings will be written
        to the output file, which most file-like objects do not
        support (default: None).
    ``outencoding``
        Overrides ``encoding`` if given.

    """

    #: Full name for the formatter, in human-readable form.
    name = None

    #: A list of short, unique identifiers that can be used to lookup
    #: the formatter from a list, e.g. using :func:`.get_formatter_by_name()`.
    aliases = []

    #: A list of fnmatch patterns that match filenames for which this
    #: formatter can produce output. The patterns in this list should be unique
    #: among all formatters.
    filenames = []

    #: If True, this formatter outputs Unicode strings when no encoding
    #: option is given.
    unicodeoutput = True

    def __init__(self, **options):
        """
        As with lexers, this constructor takes arbitrary optional arguments,
        and if you override it, you should first process your own options, then
        call the base class implementation.
        """
        self.style = _lookup_style(options.get('style', 'default'))
        self.full = get_bool_opt(options, 'full', False)
        self.title = options.get('title', '')
        self.encoding = options.get('encoding', None) or None
        if self.encoding in ('guess', 'chardet'):
            # can happen for e.g. pygmentize -O encoding=guess
            self.encoding = 'utf-8'
        self.encoding = options.get('outencoding') or self.encoding
        self.options = options

    def get_style_defs(self, arg=''):
        """
        This method must return statements or declarations suitable to define
        the current style for subsequent highlighted text (e.g. CSS classes
        in the `HTMLFormatter`).

        The optional argument `arg` can be used to modify the generation and
        is formatter dependent (it is standardized because it can be given on
        the command line).

        This method is called by the ``-S`` :doc:`command-line option <cmdline>`,
        the `arg` is then given by the ``-a`` option.
        """
        return ''

    def format(self, tokensource, outfile):
        """
        This method must format the tokens from the `tokensource` iterable and
        write the formatted version to the file object `outfile`.

        Formatter options can control how exactly the tokens are converted.
        """
        if self.encoding:
            # wrap the outfile in a StreamWriter
            outfile = codecs.lookup(self.encoding)[3](outfile)
        return self.format_unencoded(tokensource, outfile)



"""
    pygments.styles.dracula
    ~~~~~~~~~~~~~~~~~~~~~~~

    Pygments version of `Dracula` from https://github.com/dracula/dracula-theme.

    Based on the Dracula Theme for pygments by Chris Bracco.
    See https://github.com/dracula/pygments/tree/fee9ed5613d1086bc01b9d0a5a0e9867a009f571

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, Literal, \
    Number, Operator, Other, Punctuation, Text, Generic, Whitespace


class DraculaStyle(Style):

    background_color = "#282a36"
    highlight_color = "#44475a"
    line_number_color = "#f1fa8c"
    line_number_background_color = "#44475a"
    line_number_special_color = "#50fa7b"
    line_number_special_background_color = "#6272a4"

    styles = {
        Whitespace: "#f8f8f2",

        Comment: "#6272a4",
        Comment.Hashbang: "#6272a4",
        Comment.Multiline: "#6272a4",
        Comment.Preproc: "#ff79c6",
        Comment.Single: "#6272a4",
        Comment.Special: "#6272a4",

        Generic: "#f8f8f2",
        Generic.Deleted: "#8b080b",
        Generic.Emph: "#f8f8f2 underline",
        Generic.Error: "#f8f8f2",
        Generic.Heading: "#f8f8f2 bold",
        Generic.Inserted: "#f8f8f2 bold",
        Generic.Output: "#44475a",
        Generic.Prompt: "#f8f8f2",
        Generic.Strong: "#f8f8f2",
        Generic.EmphStrong: "#f8f8f2 underline",
        Generic.Subheading: "#f8f8f2 bold",
        Generic.Traceback: "#f8f8f2",

        Error: "#f8f8f2",
        Keyword: "#ff79c6",
        Keyword.Constant: "#ff79c6",
        Keyword.Declaration: "#8be9fd italic",
        Keyword.Namespace: "#ff79c6",
        Keyword.Pseudo: "#ff79c6",
        Keyword.Reserved: "#ff79c6",
        Keyword.Type: "#8be9fd",
        Literal: "#f8f8f2",
        Literal.Date: "#f8f8f2",
        Name: "#f8f8f2",
        Name.Attribute: "#50fa7b",
        Name.Builtin: "#8be9fd italic",
        Name.Builtin.Pseudo: "#f8f8f2",
        Name.Class: "#50fa7b",
        Name.Constant: "#f8f8f2",
        Name.Decorator: "#f8f8f2",
        Name.Entity: "#f8f8f2",
        Name.Exception: "#f8f8f2",
        Name.Function: "#50fa7b",
        Name.Label: "#8be9fd italic",
        Name.Namespace: "#f8f8f2",
        Name.Other: "#f8f8f2",
        Name.Tag: "#ff79c6",
        Name.Variable: "#8be9fd italic",
        Name.Variable.Class: "#8be9fd italic",
        Name.Variable.Global: "#8be9fd italic",
        Name.Variable.Instance: "#8be9fd italic",
        Number: "#ffb86c",
        Number.Bin: "#ffb86c",
        Number.Float: "#ffb86c",
        Number.Hex: "#ffb86c",
        Number.Integer: "#ffb86c",
        Number.Integer.Long: "#ffb86c",
        Number.Oct: "#ffb86c",
        Operator: "#ff79c6",
        Operator.Word: "#ff79c6",
        Other: "#f8f8f2",
        Punctuation: "#f8f8f2",
        String: "#bd93f9",
        String.Backtick: "#bd93f9",
        String.Char: "#bd93f9",
        String.Doc: "#bd93f9",
        String.Double: "#bd93f9",
        String.Escape: "#bd93f9",
        String.Heredoc: "#bd93f9",
        String.Interpol: "#bd93f9",
        String.Other: "#bd93f9",
        String.Regex: "#bd93f9",
        String.Single: "#bd93f9",
        String.Symbol: "#bd93f9",
        Text: "#f8f8f2",
    }



"""
    pygments.lexers.apdlexer
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Lexers for ANSYS Parametric Design Language.

    :copyright: Copyright 2006-2023 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re

from pygments.lexer import RegexLexer, include, words, default
from pygments.token import Comment, Keyword, Name, Number, Operator, \
    String, Generic, Punctuation, Whitespace, Escape

__all__ = ['apdlexer']


class apdlexer(RegexLexer):
    """
    For APDL source code.

    .. versionadded:: 2.9
    """
    name = 'ANSYS parametric design language'
    aliases = ['ansys', 'apdl']
    filenames = ['*.ans']
    flags = re.IGNORECASE

    # list of elements
    elafunb = ("SURF152", "SURF153", "SURF154", "SURF156", "SHELL157",
               "SURF159", "LINK160", "BEAM161", "PLANE162",
               "SHELL163", "SOLID164", "COMBI165", "MASS166",
               "LINK167", "SOLID168", "TARGE169", "TARGE170",
               "CONTA171", "CONTA172", "CONTA173", "CONTA174",
               "CONTA175", "CONTA176", "CONTA177", "CONTA178",
               "PRETS179", "LINK180", "SHELL181", "PLANE182",
               "PLANE183", "MPC184", "SOLID185", "SOLID186",
               "SOLID187", "BEAM188", "BEAM189", "SOLSH190",
               "INTER192", "INTER193", "INTER194", "INTER195",
               "MESH200", "FOLLW201", "INTER202", "INTER203",
               "INTER204", "INTER205", "SHELL208", "SHELL209",
               "CPT212", "CPT213", "COMBI214", "CPT215", "CPT216",
               "CPT217", "FLUID220", "FLUID221", "PLANE223",
               "SOLID226", "SOLID227", "PLANE230", "SOLID231",
               "SOLID232", "PLANE233", "SOLID236", "SOLID237",
               "PLANE238", "SOLID239", "SOLID240", "HSFLD241",
               "HSFLD242", "SURF251", "SURF252", "REINF263",
               "REINF264", "REINF265", "SOLID272", "SOLID273",
               "SOLID278", "SOLID279", "SHELL281", "SOLID285",
               "PIPE288", "PIPE289", "ELBOW290", "USER300", "BEAM3",
               "BEAM4", "BEAM23", "BEAM24", "BEAM44", "BEAM54",
               "COMBIN7", "FLUID79", "FLUID80", "FLUID81", "FLUID141",
               "FLUID142", "INFIN9", "INFIN47", "PLANE13", "PLANE25",
               "PLANE42", "PLANE53", "PLANE67", "PLANE82", "PLANE83",
               "PLANE145", "PLANE146", "CONTAC12", "CONTAC52",
               "LINK1", "LINK8", "LINK10", "LINK32", "PIPE16",
               "PIPE17", "PIPE18", "PIPE20", "PIPE59", "PIPE60",
               "SHELL41", "SHELL43", "SHELL57", "SHELL63", "SHELL91",
               "SHELL93", "SHELL99", "SHELL150", "SOLID5", "SOLID45",
               "SOLID46", "SOLID65", "SOLID69", "SOLID92", "SOLID95",
               "SOLID117", "SOLID127", "SOLID128", "SOLID147",
               "SOLID148", "SOLID191", "VISCO88", "VISCO89",
               "VISCO106", "VISCO107", "VISCO108", "TRANS109")

    elafunc = ("PGRAPH", "/VT", "VTIN", "VTRFIL", "VTTEMP", "PGRSET",
               "VTCLR", "VTMETH", "VTRSLT", "VTVMOD", "PGSELE",
               "VTDISC", "VTMP", "VTSEC", "PGWRITE", "VTEVAL", "VTOP",
               "VTSFE", "POUTRES", "VTFREQ", "VTPOST", "VTSL",
               "FLDATA1-40", "HFPCSWP", "MSDATA", "MSVARY", "QFACT",
               "FLOCHECK", "HFPOWER", "MSMASS", "PERI", "SPADP",
               "FLREAD", "HFPORT", "MSMETH", "PLFSS", "SPARM",
               "FLOTRAN", "HFSCAT", "MSMIR", "PLSCH", "SPFSS",
               "HFADP", "ICE", "MSNOMF", "PLSYZ", "SPICE", "HFARRAY",
               "ICEDELE", "MSPROP", "PLTD", "SPSCAN", "HFDEEM",
               "ICELIST", "MSQUAD", "PLTLINE", "SPSWP", "HFEIGOPT",
               "ICVFRC", "MSRELAX", "PLVFRC", "HFEREFINE", "LPRT",
               "MSSOLU", "/PICE", "HFMODPRT", "MSADV", "MSSPEC",
               "PLWAVE", "HFPA", "MSCAP", "MSTERM", "PRSYZ")

    elafund = ("*VOPER", "VOVLAP", "*VPLOT", "VPLOT", "VPTN", "*VPUT",
               "VPUT", "*VREAD", "VROTAT", "VSBA", "VSBV", "VSBW",
               "/VSCALE", "*VSCFUN", "VSEL", "VSLA", "*VSTAT", "VSUM",
               "VSWEEP", "VSYMM", "VTRAN", "VTYPE", "/VUP", "*VWRITE",
               "/WAIT", "WAVES", "WERASE", "WFRONT", "/WINDOW",
               "WMID", "WMORE", "WPAVE", "WPCSYS", "WPLANE", "WPOFFS",
               "WPROTA", "WPSTYL", "WRFULL", "WRITE", "WRITEMAP",
               "*WRK", "WSORT", "WSPRINGS", "WSTART", "WTBCREATE",
               "XFDATA", "XFENRICH", "XFLIST", "/XFRM", "/XRANGE",
               "XVAR", "/YRANGE", "/ZOOM", "/WB", "XMLO", "/XML",
               "CNTR", "EBLOCK", "CMBLOCK", "NBLOCK", "/TRACK",
               "CWZPLOT", "~EUI", "NELE", "EALL", "NALL", "FLITEM",
               "LSLN", "PSOLVE", "ASLN", "/VERIFY", "/SSS", "~CFIN",
               "*EVAL", "*MOONEY", "/RUNSTAT", "ALPFILL",
               "ARCOLLAPSE", "ARDETACH", "ARFILL", "ARMERGE",
               "ARSPLIT", "FIPLOT", "GAPFINISH", "GAPLIST",
               "GAPMERGE", "GAPOPT", "GAPPLOT", "LNCOLLAPSE",
               "LNDETACH", "LNFILL", "LNMERGE", "LNSPLIT", "PCONV",
               "PLCONV", "PEMOPTS", "PEXCLUDE", "PINCLUDE", "PMETH",
               "/PMETH", "PMOPTS", "PPLOT", "PPRANGE", "PRCONV",
               "PRECISION", "RALL", "RFILSZ", "RITER", "RMEMRY",
               "RSPEED", "RSTAT", "RTIMST", "/RUNST", "RWFRNT",
               "SARPLOT", "SHSD", "SLPPLOT", "SLSPLOT", "VCVFILL",
               "/OPT", "OPEQN", "OPFACT", "OPFRST", "OPGRAD",
               "OPKEEP", "OPLOOP", "OPPRNT", "OPRAND", "OPSUBP",
               "OPSWEEP", "OPTYPE", "OPUSER", "OPVAR", "OPADD",
               "OPCLR", "OPDEL", "OPMAKE", "OPSEL", "OPANL", "OPDATA",
               "OPRESU", "OPSAVE", "OPEXE", "OPLFA", "OPLGR",
               "OPLIST", "OPLSW", "OPRFA", "OPRGR", "OPRSW",
               "PILECALC", "PILEDISPSET", "PILEGEN", "PILELOAD",
               "PILEMASS", "PILERUN", "PILESEL", "PILESTIF",
               "PLVAROPT", "PRVAROPT", "TOCOMP", "TODEF", "TOFREQ",
               "TOTYPE", "TOVAR", "TOEXE", "TOLOOP", "TOGRAPH",
               "TOLIST", "TOPLOT", "TOPRINT", "TOSTAT", "TZAMESH",
               "TZDELE", "TZEGEN", "XVAROPT", "PGSAVE", "SOLCONTROL",
               "TOTAL", "VTGEOM", "VTREAL", "VTSTAT")

    elafune = ("/ANUM", "AOFFST", "AOVLAP", "APLOT", "APPEND", "APTN",
               "ARCLEN", "ARCTRM", "AREAS", "AREFINE", "AREMESH",
               "AREVERSE", "AROTAT", "ARSCALE", "ARSYM", "ASBA",
               "ASBL", "ASBV", "ASBW", "ASCRES", "ASEL", "ASIFILE",
               "*ASK", "ASKIN", "ASLL", "ASLV", "ASOL", "/ASSIGN",
               "ASUB", "ASUM", "ATAN", "ATRAN", "ATYPE", "/AUTO",
               "AUTOTS", "/AUX2", "/AUX3", "/AUX12", "/AUX15",
               "AVPRIN", "AVRES", "AWAVE", "/AXLAB", "*AXPY",
               "/BATCH", "BCSOPTION", "BETAD", "BF", "BFA", "BFADELE",
               "BFALIST", "BFCUM", "BFDELE", "BFE", "BFECUM",
               "BFEDELE", "BFELIST", "BFESCAL", "BFINT", "BFK",
               "BFKDELE", "BFKLIST", "BFL", "BFLDELE", "BFLIST",
               "BFLLIST", "BFSCALE", "BFTRAN", "BFUNIF", "BFV",
               "BFVDELE", "BFVLIST", "BIOOPT", "BIOT", "BLC4", "BLC5",
               "BLOCK", "BOOL", "BOPTN", "BSAX", "BSMD", "BSM1",
               "BSM2", "BSPLIN", "BSS1", "BSS2", "BSTE", "BSTQ",
               "BTOL", "BUCOPT", "C", "CALC", "CAMPBELL", "CBDOF",
               "CBMD", "CBMX", "CBTE", "CBTMP", "CDOPT", "CDREAD",
               "CDWRITE", "CE", "CECHECK", "CECMOD", "CECYC",
               "CEDELE", "CEINTF", "CELIST", "CENTER", "CEQN",
               "CERIG", "CESGEN", "CFACT", "*CFCLOS", "*CFOPEN",
               "*CFWRITE", "/CFORMAT", "CGLOC", "CGOMGA", "CGROW",
               "CHECK", "CHKMSH", "CINT", "CIRCLE", "CISOL",
               "/CLABEL", "/CLEAR", "CLOCAL", "CLOG", "/CLOG",
               "CLRMSHLN", "CM", "CMACEL", "/CMAP", "CMATRIX",
               "CMDELE", "CMDOMEGA", "CMEDIT", "CMGRP", "CMLIST",
               "CMMOD", "CMOMEGA", "CMPLOT", "CMROTATE", "CMSEL",
               "CMSFILE", "CMSOPT", "CMWRITE", "CNCHECK", "CNKMOD",
               "CNTR", "CNVTOL", "/COLOR", "*COMP", "COMBINE",
               "COMPRESS", "CON4", "CONE", "/CONFIG", "CONJUG",
               "/CONTOUR", "/COPY", "CORIOLIS", "COUPLE", "COVAL",
               "CP", "CPCYC", "CPDELE", "CPINTF", "/CPLANE", "CPLGEN",
               "CPLIST", "CPMERGE", "CPNGEN", "CPSGEN", "CQC",
               "*CREATE", "CRPLIM", "CS", "CSCIR", "CSDELE", "CSKP",
               "CSLIST", "CSWPLA", "CSYS", "/CTYPE", "CURR2D",
               "CUTCONTROL", "/CVAL", "CVAR", "/CWD", "CYCCALC",
               "/CYCEXPAND", "CYCFILES", "CYCFREQ", "*CYCLE",
               "CYCLIC", "CYCOPT", "CYCPHASE", "CYCSPEC", "CYL4",
               "CYL5", "CYLIND", "CZDEL", "CZMESH", "D", "DA",
               "DADELE", "DALIST", "DAMORPH", "DATA", "DATADEF",
               "DCGOMG", "DCUM", "DCVSWP", "DDASPEC", "DDELE",
               "DDOPTION", "DEACT", "DEFINE", "*DEL", "DELETE",
               "/DELETE", "DELTIM", "DELTIME", "DEMORPH", "DERIV", "DESIZE",
               "DESOL", "DETAB", "/DEVDISP", "/DEVICE", "/DFLAB",
               "DFLX", "DFSWAVE", "DIG", "DIGIT", "*DIM",
               "/DIRECTORY", "DISPLAY", "/DIST", "DJ", "DJDELE",
               "DJLIST", "DK", "DKDELE", "DKLIST", "DL", "DLDELE",
               "DLIST", "DLLIST", "*DMAT", "DMOVE", "DMPEXT",
               "DMPOPTION", "DMPRAT", "DMPSTR", "DNSOL", "*DO", "DOF",
               "DOFSEL", "DOMEGA", "*DOT", "*DOWHILE", "DSCALE",
               "/DSCALE", "DSET", "DSPOPTION", "DSUM", "DSURF",
               "DSYM", "DSYS", "DTRAN", "DUMP", "/DV3D", "DVAL",
               "DVMORPH", "DYNOPT", "E", "EALIVE", "EDADAPT", "EDALE",
               "EDASMP", "EDBOUND", "EDBX", "EDBVIS", "EDCADAPT",
               "EDCGEN", "EDCLIST", "EDCMORE", "EDCNSTR", "EDCONTACT",
               "EDCPU", "EDCRB", "EDCSC", "EDCTS", "EDCURVE",
               "EDDAMP", "EDDBL", "EDDC", "EDDRELAX", "EDDUMP",
               "EDELE", "EDENERGY", "EDFPLOT", "EDGCALE", "/EDGE",
               "EDHGLS", "EDHIST", "EDHTIME", "EDINT", "EDIPART",
               "EDIS", "EDLCS", "EDLOAD", "EDMP", "EDNB", "EDNDTSD",
               "EDNROT", "EDOPT", "EDOUT", "EDPART", "EDPC", "EDPL",
               "EDPVEL", "EDRC", "EDRD", "EDREAD", "EDRI", "EDRST",
               "EDRUN", "EDSHELL", "EDSOLV", "EDSP", "EDSTART",
               "EDTERM", "EDTP", "EDVEL", "EDWELD", "EDWRITE",
               "EEXTRUDE", "/EFACET", "EGEN", "*EIGEN", "EINFIN",
               "EINTF", "EKILL", "ELBOW", "ELEM", "ELIST", "*ELSE",
               "*ELSEIF", "EMAGERR", "EMATWRITE", "EMF", "EMFT",
               "EMID", "EMIS", "EMODIF", "EMORE", "EMSYM", "EMTGEN",
               "EMUNIT", "EN", "*END", "*ENDDO", "*ENDIF",
               "ENDRELEASE", "ENERSOL", "ENGEN", "ENORM", "ENSYM",
               "EORIENT", "EPLOT", "EQSLV", "ERASE", "/ERASE",
               "EREAD", "EREFINE", "EREINF", "ERESX", "ERNORM",
               "ERRANG", "ESCHECK", "ESEL", "/ESHAPE", "ESIZE",
               "ESLA", "ESLL", "ESLN", "ESLV", "ESOL", "ESORT",
               "ESSOLV", "ESTIF", "ESURF", "ESYM", "ESYS", "ET",
               "ETABLE", "ETCHG", "ETCONTROL", "ETDELE", "ETLIST",
               "ETYPE", "EUSORT", "EWRITE", "*EXIT", "/EXIT", "EXP",
               "EXPAND", "/EXPAND", "EXPASS", "*EXPORT", "EXPROFILE",
               "EXPSOL", "EXTOPT", "EXTREM", "EXUNIT", "F", "/FACET",
               "FATIGUE", "FC", "FCCHECK", "FCDELE", "FCLIST", "FCUM",
               "FCTYP", "FDELE", "/FDELE", "FE", "FEBODY", "FECONS",
               "FEFOR", "FELIST", "FESURF", "*FFT", "FILE",
               "FILEAUX2", "FILEAUX3", "FILEDISP", "FILL", "FILLDATA",
               "/FILNAME", "FINISH", "FITEM", "FJ", "FJDELE",
               "FJLIST", "FK", "FKDELE", "FKLIST", "FL", "FLIST",
               "FLLIST", "FLST", "FLUXV", "FLUREAD", "FMAGBC",
               "FMAGSUM", "/FOCUS", "FOR2D", "FORCE", "FORM",
               "/FORMAT", "FP", "FPLIST", "*FREE", "FREQ", "FRQSCL",
               "FS", "FSCALE", "FSDELE", "FSLIST", "FSNODE", "FSPLOT",
               "FSSECT", "FSSPARM", "FSUM", "FTCALC", "FTRAN",
               "FTSIZE", "FTWRITE", "FTYPE", "FVMESH", "GAP", "GAPF",
               "GAUGE", "GCDEF", "GCGEN", "/GCMD", "/GCOLUMN",
               "GENOPT", "GEOM", "GEOMETRY", "*GET", "/GFILE",
               "/GFORMAT", "/GLINE", "/GMARKER", "GMATRIX", "GMFACE",
               "*GO", "/GO", "/GOLIST", "/GOPR", "GP", "GPDELE",
               "GPLIST", "GPLOT", "/GRAPHICS", "/GRESUME", "/GRID",
               "/GROPT", "GRP", "/GRTYP", "/GSAVE", "GSBDATA",
               "GSGDATA", "GSLIST", "GSSOL", "/GST", "GSUM", "/GTHK",
               "/GTYPE", "HARFRQ", "/HBC", "HBMAT", "/HEADER", "HELP",
               "HELPDISP", "HEMIOPT", "HFANG", "HFSYM", "HMAGSOLV",
               "HPGL", "HPTCREATE", "HPTDELETE", "HRCPLX", "HREXP",
               "HROPT", "HROCEAN", "HROUT", "IC", "ICDELE", "ICLIST",
               "/ICLWID", "/ICSCALE", "*IF", "IGESIN", "IGESOUT",
               "/IMAGE", "IMAGIN", "IMESH", "IMMED", "IMPD",
               "INISTATE", "*INIT", "/INPUT", "/INQUIRE", "INRES",
               "INRTIA", "INT1", "INTSRF", "IOPTN", "IRLF", "IRLIST",
               "*ITENGINE", "JPEG", "JSOL", "K", "KATT", "KBC",
               "KBETW", "KCALC", "KCENTER", "KCLEAR", "KDELE",
               "KDIST", "KEEP", "KESIZE", "KEYOPT", "KEYPTS", "KEYW",
               "KFILL", "KGEN", "KL", "KLIST", "KMESH", "KMODIF",
               "KMOVE", "KNODE", "KPLOT", "KPSCALE", "KREFINE",
               "KSCALE", "KSCON", "KSEL", "KSLL", "KSLN", "KSUM",
               "KSYMM", "KTRAN", "KUSE", "KWPAVE", "KWPLAN", "L",
               "L2ANG", "L2TAN", "LANG", "LARC", "/LARC", "LAREA",
               "LARGE", "LATT", "LAYER", "LAYERP26", "LAYLIST",
               "LAYPLOT", "LCABS", "LCASE", "LCCALC", "LCCAT",
               "LCDEF", "LCFACT", "LCFILE", "LCLEAR", "LCOMB",
               "LCOPER", "LCSEL", "LCSL", "LCSUM", "LCWRITE",
               "LCZERO", "LDELE", "LDIV", "LDRAG", "LDREAD", "LESIZE",
               "LEXTND", "LFILLT", "LFSURF", "LGEN", "LGLUE",
               "LGWRITE", "/LIGHT", "LINA", "LINE", "/LINE", "LINES",
               "LINL", "LINP", "LINV", "LIST", "*LIST", "LLIST",
               "LMATRIX", "LMESH", "LNSRCH", "LOCAL", "LOVLAP",
               "LPLOT", "LPTN", "LREFINE", "LREVERSE", "LROTAT",
               "LSBA", "*LSBAC", "LSBL", "LSBV", "LSBW", "LSCLEAR",
               "LSDELE", "*LSDUMP", "LSEL", "*LSENGINE", "*LSFACTOR",
               "LSLA", "LSLK", "LSOPER", "/LSPEC", "LSREAD",
               "*LSRESTORE", "LSSCALE", "LSSOLVE", "LSTR", "LSUM",
               "LSWRITE", "/LSYMBOL", "LSYMM", "LTAN", "LTRAN",
               "LUMPM", "LVSCALE", "LWPLAN", "M", "MADAPT", "MAGOPT",
               "MAGSOLV", "/MAIL", "MAP", "/MAP", "MAP2DTO3D",
               "MAPSOLVE", "MAPVAR", "MASTER", "MAT", "MATER",
               "MCHECK", "MDAMP", "MDELE", "MDPLOT", "MEMM", "/MENU",
               "MESHING", "MFANALYSIS", "MFBUCKET", "MFCALC", "MFCI",
               "MFCLEAR", "MFCMMAND", "MFCONV", "MFDTIME", "MFELEM",
               "MFEM", "MFEXTER", "MFFNAME", "MFFR", "MFIMPORT",
               "MFINTER", "MFITER", "MFLCOMM", "MFLIST", "MFMAP",
               "MFORDER", "MFOUTPUT", "*MFOURI", "MFPSIMUL", "MFRC",
               "MFRELAX", "MFRSTART", "MFSORDER", "MFSURFACE",
               "MFTIME", "MFTOL", "*MFUN", "MFVOLUME", "MFWRITE",
               "MGEN", "MIDTOL", "/MKDIR", "MLIST", "MMASS", "MMF",
               "MODCONT", "MODE", "MODIFY", "MODMSH", "MODSELOPTION",
               "MODOPT", "MONITOR", "*MOPER", "MOPT", "MORPH", "MOVE",
               "MP", "MPAMOD", "MPCHG", "MPCOPY", "MPDATA", "MPDELE",
               "MPDRES", "/MPLIB", "MPLIST", "MPPLOT", "MPREAD",
               "MPRINT", "MPTEMP", "MPTGEN", "MPTRES", "MPWRITE",
               "/MREP", "MSAVE", "*MSG", "MSHAPE", "MSHCOPY",
               "MSHKEY", "MSHMID", "MSHPATTERN", "MSOLVE", "/MSTART",
               "MSTOLE", "*MULT", "*MWRITE", "MXPAND", "N", "NANG",
               "NAXIS", "NCNV", "NDELE", "NDIST", "NDSURF", "NEQIT",
               "/NERR", "NFORCE", "NGEN", "NKPT", "NLADAPTIVE",
               "NLDIAG", "NLDPOST", "NLGEOM", "NLHIST", "NLIST",
               "NLMESH", "NLOG", "NLOPT", "NMODIF", "NOCOLOR",
               "NODES", "/NOERASE", "/NOLIST", "NOOFFSET", "NOORDER",
               "/NOPR", "NORA", "NORL", "/NORMAL", "NPLOT", "NPRINT",
               "NREAD", "NREFINE", "NRLSUM", "*NRM", "NROPT",
               "NROTAT", "NRRANG", "NSCALE", "NSEL", "NSLA", "NSLE",
               "NSLK", "NSLL", "NSLV", "NSMOOTH", "NSOL", "NSORT",
               "NSTORE", "NSUBST", "NSVR", "NSYM", "/NUMBER",
               "NUMCMP", "NUMEXP", "NUMMRG", "NUMOFF", "NUMSTR",
               "NUMVAR", "NUSORT", "NWPAVE", "NWPLAN", "NWRITE",
               "OCDATA", "OCDELETE", "OCLIST", "OCREAD", "OCTABLE",
               "OCTYPE", "OCZONE", "OMEGA", "OPERATE", "OPNCONTROL",
               "OUTAERO", "OUTOPT", "OUTPR", "/OUTPUT", "OUTRES",
               "OVCHECK", "PADELE", "/PAGE", "PAGET", "PAPUT",
               "PARESU", "PARTSEL", "PARRES", "PARSAV", "PASAVE",
               "PATH", "PAUSE", "/PBC", "/PBF", "PCALC", "PCGOPT",
               "PCIRC", "/PCIRCLE", "/PCOPY", "PCROSS", "PDANL",
               "PDCDF", "PDCFLD", "PDCLR", "PDCMAT", "PDCORR",
               "PDDMCS", "PDDOEL", "PDEF", "PDEXE", "PDHIST",
               "PDINQR", "PDLHS", "PDMETH", "PDOT", "PDPINV",
               "PDPLOT", "PDPROB", "PDRESU", "PDROPT", "/PDS",
               "PDSAVE", "PDSCAT", "PDSENS", "PDSHIS", "PDUSER",
               "PDVAR", "PDWRITE", "PERBC2D", "PERTURB", "PFACT",
               "PHYSICS", "PIVCHECK", "PLCAMP", "PLCFREQ", "PLCHIST",
               "PLCINT", "PLCPLX", "PLCRACK", "PLDISP", "PLESOL",
               "PLETAB", "PLFAR", "PLF2D", "PLGEOM", "PLLS", "PLMAP",
               "PLMC", "PLNEAR", "PLNSOL", "/PLOPTS", "PLORB", "PLOT",
               "PLOTTING", "PLPAGM", "PLPATH", "PLSECT", "PLST",
               "PLTIME", "PLTRAC", "PLVAR", "PLVECT", "PLZZ",
               "/PMACRO", "PMAP", "PMGTRAN", "PMLOPT", "PMLSIZE",
               "/PMORE", "PNGR", "/PNUM", "POINT", "POLY", "/POLYGON",
               "/POST1", "/POST26", "POWERH", "PPATH", "PRANGE",
               "PRAS", "PRCAMP", "PRCINT", "PRCPLX", "PRED",
               "PRENERGY", "/PREP7", "PRERR", "PRESOL", "PRETAB",
               "PRFAR", "PRI2", "PRIM", "PRINT", "*PRINT", "PRISM",
               "PRITER", "PRJSOL", "PRNEAR", "PRNLD", "PRNSOL",
               "PROD", "PRORB", "PRPATH", "PRRFOR", "PRRSOL",
               "PRSCONTROL", "PRSECT", "PRTIME", "PRVAR", "PRVECT",
               "PSCONTROL", "PSCR", "PSDCOM", "PSDFRQ", "PSDGRAPH",
               "PSDRES", "PSDSPL", "PSDUNIT", "PSDVAL", "PSDWAV",
               "/PSEARCH", "PSEL", "/PSF", "PSMAT", "PSMESH",
               "/PSPEC", "/PSTATUS", "PSTRES", "/PSYMB", "PTR",
               "PTXY", "PVECT", "/PWEDGE", "QDVAL", "QRDOPT", "QSOPT",
               "QUAD", "/QUIT", "QUOT", "R", "RACE", "RADOPT",
               "RAPPND", "RATE", "/RATIO", "RBE3", "RCON", "RCYC",
               "RDEC", "RDELE", "READ", "REAL", "REALVAR", "RECTNG",
               "REMESH", "/RENAME", "REORDER", "*REPEAT", "/REPLOT",
               "RESCOMBINE", "RESCONTROL", "RESET", "/RESET", "RESP",
               "RESUME", "RESVEC", "RESWRITE", "*RETURN", "REXPORT",
               "REZONE", "RFORCE", "/RGB", "RIGID", "RIGRESP",
               "RIMPORT", "RLIST", "RMALIST", "RMANL", "RMASTER",
               "RMCAP", "RMCLIST", "/RMDIR", "RMFLVEC", "RMLVSCALE",
               "RMMLIST", "RMMRANGE", "RMMSELECT", "RMNDISP",
               "RMNEVEC", "RMODIF", "RMORE", "RMPORDER", "RMRESUME",
               "RMRGENERATE", "RMROPTIONS", "RMRPLOT", "RMRSTATUS",
               "RMSAVE", "RMSMPLE", "RMUSE", "RMXPORT", "ROCK",
               "ROSE", "RPOLY", "RPR4", "RPRISM", "RPSD", "RSFIT",
               "RSOPT", "RSPLIT", "RSPLOT", "RSPRNT", "RSSIMS",
               "RSTMAC", "RSTOFF", "RSURF", "RSYMM", "RSYS", "RTHICK",
               "SABS", "SADD", "SALLOW", "SAVE", "SBCLIST", "SBCTRAN",
               "SDELETE", "SE", "SECCONTROL", "SECDATA",
               "SECFUNCTION", "SECJOINT", "/SECLIB", "SECLOCK",
               "SECMODIF", "SECNUM", "SECOFFSET", "SECPLOT",
               "SECREAD", "SECSTOP", "SECTYPE", "SECWRITE", "SED",
               "SEDLIST", "SEEXP", "/SEG", "SEGEN", "SELIST", "SELM",
               "SELTOL", "SENERGY", "SEOPT", "SESYMM", "*SET", "SET",
               "SETFGAP", "SETRAN", "SEXP", "SF", "SFA", "SFACT",
               "SFADELE", "SFALIST", "SFBEAM", "SFCALC", "SFCUM",
               "SFDELE", "SFE", "SFEDELE", "SFELIST", "SFFUN",
               "SFGRAD", "SFL", "SFLDELE", "SFLEX", "SFLIST",
               "SFLLIST", "SFSCALE", "SFTRAN", "/SHADE", "SHELL",
               "/SHOW", "/SHOWDISP", "SHPP", "/SHRINK", "SLIST",
               "SLOAD", "SMALL", "*SMAT", "SMAX", "/SMBC", "SMBODY",
               "SMCONS", "SMFOR", "SMIN", "SMOOTH", "SMRTSIZE",
               "SMSURF", "SMULT", "SNOPTION", "SOLU", "/SOLU",
               "SOLUOPT", "SOLVE", "SORT", "SOURCE", "SPACE",
               "SPCNOD", "SPCTEMP", "SPDAMP", "SPEC", "SPFREQ",
               "SPGRAPH", "SPH4", "SPH5", "SPHERE", "SPLINE", "SPLOT",
               "SPMWRITE", "SPOINT", "SPOPT", "SPREAD", "SPTOPT",
               "SPOWER", "SPUNIT", "SPVAL", "SQRT", "*SREAD", "SRSS",
               "SSBT", "/SSCALE", "SSLN", "SSMT", "SSPA", "SSPB",
               "SSPD", "SSPE", "SSPM", "SSUM", "SSTATE", "STABILIZE",
               "STAOPT", "STAT", "*STATUS", "/STATUS", "STEF",
               "STORE", "SUBOPT", "SUBSET", "SUCALC",
               "SUCR", "SUDEL", "SUEVAL", "SUGET", "SUMAP", "SUMTYPE",
               "SUPL", "SUPR", "SURESU", "SUSAVE", "SUSEL", "SUVECT",
               "SV", "SVPLOT", "SVTYP", "SWADD", "SWDEL", "SWGEN",
               "SWLIST", "SYNCHRO", "/SYP", "/SYS", "TALLOW",
               "TARGET", "*TAXIS", "TB", "TBCOPY", "TBDATA", "TBDELE",
               "TBEO", "TBIN", "TBFIELD", "TBFT", "TBLE", "TBLIST",
               "TBMODIF", "TBPLOT", "TBPT", "TBTEMP", "TCHG", "/TEE",
               "TERM", "THEXPAND", "THOPT", "TIFF", "TIME",
               "TIMERANGE", "TIMINT", "TIMP", "TINTP",
               "/TLABEL", "TOFFST", "*TOPER", "TORQ2D", "TORQC2D",
               "TORQSUM", "TORUS", "TRANS", "TRANSFER", "*TREAD",
               "TREF", "/TRIAD", "/TRLCY", "TRNOPT", "TRPDEL",
               "TRPLIS", "TRPOIN", "TRTIME", "TSHAP", "/TSPEC",
               "TSRES", "TUNIF", "TVAR", "/TXTRE", "/TYPE", "TYPE",
               "/UCMD", "/UDOC", "/UI", "UIMP", "/UIS", "*ULIB", "/UPF",
               "UNDELETE", "UNDO", "/UNITS", "UNPAUSE", "UPCOORD",
               "UPGEOM", "*USE", "/USER", "USRCAL", "USRDOF",
               "USRELEM", "V", "V2DOPT", "VA", "*VABS", "VADD",
               "VARDEL", "VARNAM", "VATT", "VCLEAR", "*VCOL",
               "/VCONE", "VCROSS", "*VCUM", "VDDAM", "VDELE", "VDGL",
               "VDOT", "VDRAG", "*VEC", "*VEDIT", "VEORIENT", "VEXT",
               "*VFACT", "*VFILL", "VFOPT", "VFQUERY", "VFSM",
               "*VFUN", "VGEN", "*VGET", "VGET", "VGLUE", "/VIEW",
               "VIMP", "VINP", "VINV", "*VITRP", "*VLEN", "VLIST",
               "VLSCALE", "*VMASK", "VMESH", "VOFFST", "VOLUMES")

    # list of in-built () functions
    elafunf = ("NX()", "NY()", "NZ()", "KX()", "KY()", "KZ()", "LX()",
               "LY()", "LZ()", "LSX()", "LSY()", "LSZ()", "NODE()",
               "KP()", "DISTND()", "DISTKP()", "DISTEN()", "ANGLEN()",
               "ANGLEK()", "NNEAR()", "KNEAR()", "ENEARN()",
               "AREAND()", "AREAKP()", "ARNODE()", "NORMNX()",
               "NORMNY()", "NORMNZ()", "NORMKX()", "NORMKY()",
               "NORMKZ()", "ENEXTN()", "NELEM()", "NODEDOF()",
               "ELADJ()", "NDFACE()", "NMFACE()", "ARFACE()", "UX()",
               "UY()", "UZ()", "ROTX()", "ROTY()", "ROTZ()", "TEMP()",
               "PRES()", "VX()", "VY()", "VZ()", "ENKE()", "ENDS()",
               "VOLT()", "MAG()", "AX()", "AY()", "AZ()",
               "VIRTINQR()", "KWGET()", "VALCHR()", "VALHEX()",
               "CHRHEX()", "STRFILL()", "STRCOMP()", "STRPOS()",
               "STRLENG()", "UPCASE()", "LWCASE()", "JOIN()",
               "SPLIT()", "ABS()", "SIGN()", "CXABS()", "EXP()",
               "LOG()", "LOG10()", "SQRT()", "NINT()", "MOD()",
               "RAND()", "GDIS()", "SIN()", "COS()", "TAN()",
               "SINH()", "COSH()", "TANH()", "ASIN()", "ACOS()",
               "ATAN()", "ATAN2()")

    elafung = ("NSEL()", "ESEL()", "KSEL()", "LSEL()", "ASEL()",
               "VSEL()", "NDNEXT()", "ELNEXT()", "KPNEXT()",
               "LSNEXT()", "ARNEXT()", "VLNEXT()", "CENTRX()",
               "CENTRY()", "CENTRZ()")

    elafunh = ("~CAT5IN", "~CATIAIN", "~PARAIN", "~PROEIN", "~SATIN",
               "~UGIN", "A", "AADD", "AATT", "ABEXTRACT", "*ABBR",
               "ABBRES", "ABBSAV", "ABS", "ACCAT", "ACCOPTION",
               "ACEL", "ACLEAR", "ADAMS", "ADAPT", "ADD", "ADDAM",
               "ADELE", "ADGL", "ADRAG", "AESIZE", "AFILLT", "AFLIST",
               "AFSURF", "*AFUN", "AGEN", "AGLUE", "AINA", "AINP",
               "AINV", "AL", "ALIST", "ALLSEL", "ALPHAD", "AMAP",
               "AMESH", "/AN3D", "ANCNTR", "ANCUT", "ANCYC", "ANDATA",
               "ANDSCL", "ANDYNA", "/ANFILE", "ANFLOW", "/ANGLE",
               "ANHARM", "ANIM", "ANISOS", "ANMODE", "ANMRES",
               "/ANNOT", "ANORM", "ANPRES", "ANSOL", "ANSTOAQWA",
               "ANSTOASAS", "ANTIME", "ANTYPE")

    special = ("/COM", "/TITLE", "STITLE")

    elements = ("SOLID5",
                "LINK11",
                "PLANE13",
                "COMBIN14",
                "MASS2",
                "PLANE25",
                "MATRIX27",
                "FLUID29",
                "FLUID30",
                "LINK31",
                "LINK33",
                "LINK34",
                "PLANE35",
                "SOURC36",
                "COMBIN37",
                "FLUID38",
                "COMBIN39",
                "COMBIN40",
                "INFIN47",
                "MATRIX50",
                "PLANE55",
                "SHELL61",
                "LINK68",
                "SOLID70",
                "MASS71",
                "PLANE75",
                "PLANE77",
                "PLANE78",
                "PLANE83",
                "SOLID87",
                "SOLID90",
                "CIRCU94",
                "SOLID96",
                "SOLID98",
                "INFIN110",
                "INFIN111",
                "FLUID116",
                "PLANE121",
                "SOLID122",
                "SOLID123",
                "CIRCU124",
                "CIRCU125",
                "TRANS126",
                "FLUID129",
                "FLUID130",
                "SHELL131",
                "SHELL132",
                "FLUID136",
                "FLUID138",
                "FLUID139",
                "SURF151",
                "SURF152",
                "SURF153",
                "SURF154",
                "SURF155",
                "SURF156",
                "SHELL157",
                "SURF159",
                "TARGE169",
                "TARGE170",
                "CONTA172",
                "CONTA174",
                "CONTA175",
                "CONTA177",
                "CONTA178",
                "PRETS179",
                "LINK180",
                "SHELL181",
                "PLANE182",
                "PLANE183",
                "MPC184",
                "SOLID185",
                "SOLID186",
                "SOLID187",
                "BEAM188",
                "BEAM189",
                "SOLSH190",
                "INTER192",
                "INTER193",
                "INTER194",
                "INTER195",
                "MESH200",
                "FOLLW201",
                "INTER202",
                "INTER203",
                "INTER204",
                "INTER205",
                "SHELL208",
                "SHELL209",
                "CPT212",
                "CPT213",
                "COMBI214",
                "CPT215",
                "CPT216",
                "CPT217",
                "FLUID218",
                "FLUID220",
                "FLUID221",
                "PLANE222",
                "PLANE223",
                "SOLID225",
                "SOLID226",
                "SOLID227",
                "PLANE230",
                "SOLID231",
                "SOLID232",
                "PLANE233",
                "SOLID236",
                "SOLID237",
                "PLANE238",
                "SOLID239",
                "SOLID240",
                "HSFLD241",
                "HSFLD242",
                "COMBI250",
                "SURF251",
                "SURF252",
                "INFIN257",
                "REINF263",
                "REINF264",
                "REINF265",
                "SOLID272",
                "SOLID273",
                "SOLID278",
                "SOLID279",
                "CABLE280",
                "SHELL281",
                "SOLID285",
                "PIPE288",
                "PIPE289",
                "ELBOW290",
                "SOLID291",
                "PLANE292",
                "PLANE293",
                "USER300")

    tokens = {
        'root': [
            (r'[^\S\n]+', Whitespace),
            (words((elafunb+elafunc+elafund+elafune+elafunh+special), suffix=r'\b'), Keyword, 'non-keyword'),
            default('non-keyword'),
        ],
        'non-keyword': [
            (r'!.*\n', Comment, '#pop'),
            (r'%.*?%', Escape),
            include('strings'),
            include('nums'),
            (words((elafunf+elafung), suffix=r'\b'), Name.Builtin),
            (words((elements), suffix=r'\b'), Name.Property),
            include('core'),
            (r'AR[0-9]+', Name.Variable.Instance),
            (r'[a-z_][a-z0-9_]*', Name.Variable),
            (r'\n+', Whitespace, '#pop'),
            (r'[^\S\n]+', Whitespace),
        ],
        'core': [
            # Operators
            (r'(\*\*|\*|\+|-|\/|<|>|<=|>=|==|\/=|=|\(|\))', Operator),
            (r'/EOF', Generic.Emph),
            (r'[\.(),:&;]', Punctuation),
        ],
        'strings': [
            (r'(?s)"(\\\\|\\[0-7]+|\\.|[^"\\])*"', String.Double),
            (r"(?s)'(\\\\|\\[0-7]+|\\.|[^'\\])*'", String.Single),
            (r'[$%]', String.Symbol),
        ],
        'nums': [
            (r'[+-]?\d*\.\d+([efEF][-+]?\d+)?', Number.Float), # with dot
            (r'([+-]?\d+([efEF][-+]?\d+))', Number.Float), # With scientific notation
            (r'\b\d+(?![.ef])', Number.Integer), # integer simple
        ]
    }
